import os
import re
import tarfile
import glob

from tqdm import tqdm

from proteinshake.datasets import Dataset
from proteinshake.utils import download_url, extract_tar, load, save, unzip_file

# A map of organism names to their download file names. See https://alphafold.ebi.ac.uk/download
AF_DATASET_NAMES = {
    'arabidopsis_thaliana': 'UP000006548_3702_ARATH',
    'caenorhabditis_elegans': 'UP000001940_6239_CAEEL',
    'candida_albicans': 'UP000000559_237561_CANAL',
    'danio_rerio': 'UP000000437_7955_DANRE',
    'dictyostelium_discoideum': 'UP000002195_44689_DICDI',
    'drosophila_melanogaster': 'UP000000803_7227_DROME',
    'escherichia_coli': 'UP000000625_83333_ECOLI',
    'glycine_max': 'UP000008827_3847_SOYBN',
    'homo_sapiens': 'UP000005640_9606_HUMAN',
    'methanocaldococcus_jannaschii': 'UP000000805_243232_METJA',
    'mus_musculus': 'UP000000589_10090_MOUSE',
    'oryza_sativa': 'UP000059680_39947_ORYSJ',
    'rattus_norvegicus': 'UP000002494_10116_RAT',
    'saccharomyces_cerevisiae': 'UP000002311_559292_YEAST',
    'schizosaccharomyces_pombe': 'UP000002485_284812_SCHPO',
    'zea_mays': 'UP000007305_4577_MAIZE',
    'swissprot': 'swissprot_pdb',
}

class AlphaFoldDataset(Dataset):
    """ 3D structures predicted from sequence by AlphaFold.

    Requires the `organism` name to be specified. See https://alphafold.ebi.ac.uk/download for a full list of available organsims.
    Pass the full latin organism name separated by a space or underscore. `organism` can also be 'swissprot', in which case the full SwissProt structure predictions will be downloaded (ca. 500.000).

    .. list-table :: Data Properties
       :widths: 50 50
       :header-rows: 1

       * - organism
         - # proteins
       * - ``'arabidopsis_thaliana'``
         - 27,386
       * - ``'caenorhabditis_elegans'``
         - 19,613
       * - ``'candida_albicans'``
         - 5,951
       * - ``'danio_rerio'``
         - 24,430
       * - ``'dictyostelium_discoideum'``
         - 12,485
       * - ``'drosophila_melanogaster'``
         - 13,318
       * - ``'escherichia_coli'``
         - 4,362
       * - ``'glycine_max'``
         - 55,696
       * - ``'homo_sapiens'``
         - 23,172
       * - ``'methanocaldococcus_jannaschii'``
         - 1,773
       * - ``'mus_musculus'``
         - 21,398
       * - ``'oryza_sativa'``
         - 43,631
       * - ``'rattus_norvegicus'``
         - 21,069
       * - ``'saccharomyces_cerevisiae'``
         - 6,016
       * - ``'schizosaccharomyces_pombe'``
         - 5,104
       * - ``'zea_mays'``
         - 39,203
       * - ``'swissprot'``
         - 541,143


    Parameters
    ----------
    organism: str
        The organism name or 'swissprot'.
    """

    exlude_args_from_signature = ['organism']

    def __init__(self, organism, version='v4', only_single_chain=True, **kwargs):
        self.organism = organism.lower().replace(' ','_')
        self.base_url = 'https://ftp.ebi.ac.uk/pub/databases/alphafold/latest/'
        self.version = version
        super().__init__(only_single_chain=only_single_chain, **kwargs)

    @property
    def name(self):
        return f'{self.__class__.__name__}_{self.organism}'

    def get_raw_files(self):
        return glob.glob(f'{self.root}/raw/*/*.pdb')[:self.limit]

    def get_id_from_filename(self, filename):
        return re.search('(?<=AF-)(.*)(?=-F.+-model)', filename).group()

    def download(self):
        os.makedirs(f'{self.root}/raw/{self.organism}', exist_ok=True)
        download_url(self.base_url+AF_DATASET_NAMES[self.organism]+f'_{self.version}.tar', f'{self.root}/raw/{self.organism}')
        extract_tar(f'{self.root}/raw/{self.organism}/{AF_DATASET_NAMES[self.organism]}_{self.version}.tar', f'{self.root}/raw/{self.organism}')
        [unzip_file(f) for f in tqdm(glob.glob(f'{self.root}/raw/*/*.pdb.gz')[:self.limit], desc='Unzipping')]



# -*- coding: utf-8 -*-
import glob
import os
import re
import os.path as osp

from rdkit import Chem
from rdkit import RDLogger
from rdkit.Chem import MACCSkeys
from rdkit.Chem import AllChem

import numpy as np

from proteinshake.datasets import Dataset
from proteinshake.utils import extract_tar, download_url

RDLogger.DisableLog('rdApp.*') # disable warnings

class ProteinLigandInterfaceDataset(Dataset):
    """Proteins bound to small molecules from PDBBind with binding site, ligand and affinity information.

    Parameters
    ----------
    root: str
        Root directory where the dataset should be saved.
    name: str
        The name of the dataset.
    version: str
        PDBBind version to use.

    .. list-table:: Dataset stats
       :widths: 100
       :header-rows: 1

       * - # proteins
       * - 4642


   .. list-table:: Annotations
      :widths: 20 55 25
      :header-rows: 1

      * - Attribute
        - Key
        - Sample value
      * - Dissociation constant (kd)
        - :code:`protein['protein']['kd']`
        - :code:`77.0`
      * - Affinity
        - :code:`protein['protein']['neglog_aff']`
        - :code:`4.11000`
      * - Resolution (Angstroms)
        - :code:`protein['protein']['resolution']`
        - :code:`2.20`
      * - Year solved
        - :code:`protein['protein']['year']`
        - :code:`2016`
      * - Ligand identifier (PDB code)
        - :code:`protein['protein']['ligands_id']`
        - :code:`IEE`
      * - Ligand SMILES 
        - :code:`protein['protein']['ligand_smiles']`
        - :code:`'Cc1ccc(CNc2cc(Cl)nc(N)n2)cc1'`
      * - Molecular ingerprints
        - :code:`protein['protein']['fp_maccs']`, :code:`protein['protein']['fp_morgan_r2']`
        - :code:`'[..,0, 0, 1, 0, 1, 0, 0, 0,..]`
      * - Molecular ingerprints
        - :code:`protein['protein']['fp_maccs']`, :code:`protein['protein']['fp_morgan_r2']`
        - :code:`'[..,0, 0, 1, 0, 1, 0, 0, 0,..]`
      * - Binding site (1 if in binding site, 0 else)
        - :code:`protein['residue']['binding_site']`
        - :code:`'[..,0, 0, 1, 0, 1, 0, 0, 0,..]`
    """

    def __init__(self, version='2020', **kwargs):
        self.version = version
        super().__init__(**kwargs)

    def get_raw_files(self):
        return glob.glob(f'{self.root}/raw/files/*_protein.pdb')[:self.limit]

    def get_id_from_filename(self, filename):
        return filename[:4]

    def affinity_parse(self, s):
        """ Parse the affinity string. e.g. `Kd=30uM`.
        Parameters
        ----------
        s: str
            Affinity measurement string to parse.
        Returns
        -------
        dict
            Dictionary containing parsed affinity information. `value` key stores
            the float value of the measurement. `operator` is the logical operator
            (e.g. `=`, `>`) applied to the value, `unit` is `uM, nM, pM` and
            `measure` is the type experimental measurement (e.g. `Kd, Ki, IC50`)
        """
        operator = "".join(re.findall(r"[=|<|>|~]", s))
        measures = ['Kd', 'Ki', 'IC50']
        for m in measures:
            if s.startswith(m):
                measure = m
                break
        value = float(re.search(r"\d+[.,]?\d*", s).group())
        unit = re.search(r"[m|u|n|f|p]M", s).group()

        return {'operator': operator,
                'measure': measure,
                'value': value,
                'unit': unit
                }

    def parse_pdbbind_PL_index(self, index_path):
        """
        > INDEX_refined_data.2020
        # ==============================================================================
        # List of the protein-ligand complexes in the PDBbind refined set v.2020
        # 5316 protein-ligand complexes in total, which are ranked by binding data
        # Latest update: July 2021
        # PDB code, resolution, release year, -logKd/Ki, Kd/Ki, reference, ligand name
        # ==============================================================================
        2r58  2.00  2007   2.00  Kd=10mM       // 2r58.pdf (MLY)
        3c2f  2.35  2008   2.00  Kd=10.1mM     // 3c2f.pdf (PRP)
        3g2y  1.31  2009   2.00  Ki=10mM       // 3g2y.pdf (GF4)
        3pce  2.06  1998   2.00  Ki=10mM       // 3pce.pdf (3HP)
        4qsu  1.90  2014   2.00  Kd=10mM       // 4qsu.pdf (TDR)
        4qsv  1.90  2014   2.00  Kd=10mM       // 4qsv.pdf (THM)
        """
        data = {}
        with open(index_path, 'r') as ind_file:
            for line in ind_file:
                if line.startswith("#"):
                    continue
                pre, post = line.split("//")
                pdbid, res, date, neglog, kd = pre.split()
                kd = self.affinity_parse(kd)

                lig_id = post.split("(")[1].rstrip(")")

                # remove peptide ligands
                if lig_id.endswith('-mer'):
                    continue
                data[pdbid] = {
                    'resolution': float(res),
                    'date': int(date),
                    'kd': kd,
                    'neglog_aff': float(neglog),
                    'ligand_id': lig_id
                }
        return data

    def download(self):
        download_url(f'https://pdbbind.oss-cn-hangzhou.aliyuncs.com/download/PDBbind_v{self.version}_refined.tar.gz', f'{self.root}/raw')
        extract_tar(f'{self.root}/raw/PDBbind_v{self.version}_refined.tar.gz', f'{self.root}/raw/files', extract_members=True, strip=1)
        self.index_data = self.parse_pdbbind_PL_index(f'{self.root}/raw/files/INDEX_refined_data.{self.version}')
        
    def add_protein_attributes(self, protein):
        pocket = self.pdb2df(f'{self.root}/raw/files/{protein["protein"]["ID"]}_pocket.pdb')
        ligand = Chem.MolFromMolFile(f'{self.root}/raw/files/{protein["protein"]["ID"]}_ligand.sdf')

        if ligand is None:
            return None
        smiles = Chem.MolToSmiles(ligand)
        fp_morgan = list(map(int, AllChem.GetMorganFingerprintAsBitVect(ligand, 2, nBits=1024).ToBitString()))
        fp_maccs = list(map(int, MACCSkeys.GenMACCSKeys(ligand).ToBitString()))

        pocket_res = np.array(pocket.loc[pocket['atom_type'] == 'CA']['residue_number'].tolist())
        protein_res = np.array(protein['residue']['residue_number']).reshape(-1, 1)

        pocket_atom = np.array(pocket['residue_number'].tolist())
        protein_atom = np.array(protein['atom']['residue_number']).reshape(-1, 1)

        is_site_res = np.zeros_like(protein_res)
        is_site_atom  = np.zeros_like(protein_atom)

        is_site_res[(pocket_res == protein_res).sum(axis=1).nonzero()] = 1.
        is_site_atom[(pocket_atom == protein_atom).sum(axis=1).nonzero()] = 1.


        protein['residue']['binding_site'] = list(map(int, is_site_res.squeeze()))
        protein['atom']['binding_site'] = list(map(int, is_site_atom.squeeze()))

        bind_data = self.index_data[protein['protein']['ID']]
        protein['protein']['kd'] = bind_data['kd']['value']
        protein['protein']['neglog_aff'] = bind_data['neglog_aff']
        protein['protein']['resolution'] = bind_data['resolution']
        protein['protein']['year'] = bind_data['date']
        protein['protein']['ligand_id'] = bind_data['ligand_id']
        protein['protein']['ligand_smiles'] = smiles

        protein['protein']['fp_maccs'] = fp_maccs
        protein['protein']['fp_morgan_r2'] = fp_morgan

        return protein

    def describe(self):
        desc = super().describe()
        desc['property'] = "Small Mol. Binding Site (residue-level)"
        desc['values'] = 2
        desc['type'] = 'Binary'
        return desc



from .dataset import Dataset
from .rcsb import RCSBDataset
from .enzyme_commission import EnzymeCommissionDataset
from .gene_ontology import GeneOntologyDataset
from .protein_protein_interface import ProteinProteinInterfaceDataset
from .protein_ligand_interface import ProteinLigandInterfaceDataset
from .protein_family import ProteinFamilyDataset
from .tm_align import TMAlignDataset
from .alphafold import AlphaFoldDataset
from .scop import SCOPDataset
from .protein_ligand_decoys import ProteinLigandDecoysDataset

__all__ = [
    'Dataset',
    'RCSBDataset',
    'EnzymeCommissionDataset',
    'GeneOntologyDataset',
    'ProteinProteinInterfaceDataset',
    'ProteinLigandInterfaceDataset',
    'ProteinFamilyDataset',
    'TMAlignDataset',
    'AlphaFoldDataset',
    'SCOPDataset',
    'ProteinLigandDecoysDataset'
    ]

classes = __all__



"""
Abstract class for transforming a protein.
"""

class Transform:
    """ A callable object which accepts a protein dictionary and returns an updated version of it."""
    def __init__(self):
        pass

    def __call__(self, protein):
        """ Accepts a protein dict and returns a new protein dict

        Arguments
        ----------
        protein: dict
            A protein dictionary.

        """
        raise NotImplementedError

class IdentityTransform:
    """ Do nothing to the protein"""
    def __call__(self, protein):
        return protein

class Compose:

    def __init__(self, transforms):
        self.transforms = transforms

    def __call__(self, data):
        for transform in self.transforms:
            data = transform(data)
        return data

    def __repr__(self):
        args = [f'  {transform}' for transform in self.transforms]
        return '{}([\n{}\n])'.format(self.__class__.__name__, ',\n'.join(args))


import networkx as nx
from proteinshake.frameworks.dataset import FrameworkDataset


class NetworkxGraphDataset(FrameworkDataset):

    def convert_to_framework(self, data_item):
        nodes, adj = data_item.data
        data = nx.from_scipy_sparse_array(adj)
        data.add_nodes_from(nodes)
        return data



import pandas as pd
from joblib import Parallel, delayed

from tqdm import tqdm

from proteinshake.datasets import RCSBDataset
from proteinshake.utils import download_url

class SCOPDataset(RCSBDataset):
    """ Proteins for which the SCOP classification is known.

    .. list-table:: Dataset stats
       :widths: 100
       :header-rows: 1

       * - # proteins
       * - 10066


   .. list-table:: Annotations
      :widths: 25 45 35
      :header-rows: 1

      * - Attribute
        - Key
        - Sample value
      * - Protein type
        - :code:`protein['protein']['SCOP-TP']`
        - :code:`'1'`
      * - Protein class 
        - :code:`protein['protein']['SCOP-CL']`
        - ``'1000000'``
      * - Superfamily 
        - :code:`protein['protein']['SCOP-SF']`
        - ``'3000001'``
      * - Family
        - :code:`protein['protein']['SCOP-FA']`
        - ``'4002873'``


    """

    def _parse_scop(self, path):
        names = ['FA-DOMID', 'FA-PDBID', 'FA-PDBREG', 'FA-UNIID', 'FA-UNIREG', 'SF-DOMID', 'SF-PDBID', 'SF-PDBREG', 'SF-UNIID', 'SF-UNIREG', 'SCOPCLA']
        df = pd.read_csv(path, sep=' ', comment='#', names=names, dtype=str)
        return {k: dict([cla.split("=") for cla in v.split(",")]) for k,v in zip(df['FA-PDBID'], df['SCOPCLA'])}

    def download(self):
        # get the annots
        download_url(f'http://scop.mrc-lmb.cam.ac.uk/files/scop-cla-latest.txt', f'{self.root}/raw/scop.txt')
        self.scop = self._parse_scop(f'{self.root}/raw/scop.txt')
        ids = list(self.scop['FA-PDBID'].unique())

        # get the proteins
        if self.n_jobs == 1:
            print('Warning: Downloading an RCSB dataset with use_precompute = False is very slow. Consider increasing n_jobs.')
        ids = ids[:self.limit] # for testing

        failed = Parallel(n_jobs=self.n_jobs)(delayed(self.download_from_rcsb)(id) for id in tqdm(ids, desc='Downloading PDBs'))
        failed = [f for f in failed if not f is True]
        if len(failed)>0:
            print(f'Failed to download {len(failed)} PDB files.')

    def add_protein_attributes(self, protein):
        """ We annotate the protein with the scop classifications at each level.

        SCOPCLA - SCOP domain classification. The abbreviations denote: TP=protein type, CL=protein class, CF=fold, SF=superfamily, FA=family
        """
        protein_id = protein['protein']['ID'].upper()
        if not protein_id in self.scop: return None
        for cla, val in self.scop[protein_id].items():
            protein['protein']['SCOP-' + cla] = val
        return protein



import requests, glob, json, os, random
import pandas as pd
from tqdm import tqdm
from joblib import Parallel, delayed

from proteinshake.datasets import Dataset
from proteinshake.utils import download_url, unzip_file

class RCSBDataset(Dataset):
    """ Non-redundant structures taken from RCSB Protein Databank.

    This class also serves as a base class for all RCSB derived datasets. It can be subclassed by defining a default ``query`` argument. The query is a list of triplets ``(attribute, operator, value)`` according to `this <https://search.rcsb.org/#attribute-queries>`_ and `this <https://data.rcsb.org/data-attributes.html>`_ , which is passed to the REST API call to RCSB. See e.g. the GODataset subclass for an example. To find the right attributes, the queries can be constructed by doing an advanced search `at RCSB <https://www.rcsb.org/search/advanced>`_ and exporting to JSON. Also compare the API call in the :meth:`download()` method.

    It uses RCSB's integrated sequence similarity filtering to remove redundant proteins.

    Also, only single chain proteins are used. Change the REST payload in `download` to override this behaviour.

    Parameters
    ----------
    query: list
        A list of triplets `(attribute, operator, value)` to be added to the REST API call to RCSB.
    """

    def __init__(self, query=[], from_list=None, only_single_chain=True, max_requests=20, **kwargs):
        self.query = query
        self.from_list = from_list
        self.max_requests = max_requests
        super().__init__(only_single_chain=only_single_chain, **kwargs)

    def get_raw_files(self):
        return glob.glob(f'{self.root}/raw/files/*.pdb')

    def get_id_from_filename(self, filename):
        return filename[:4]

    def download(self):
        """ Fetches PDBs from RCSB with an API call. The default query selects protein-only
        structures with a single chain.
        """

        if self.n_jobs == 1:
            print('Warning: Downloading an RCSB dataset with use_precompute = False is very slow. Consider increasing n_jobs.')
        total = None
        i = 0
        batch_size = 5000
        ids = []
        if not self.from_list is None:
            ids = self.from_list
        else:
            while total is None or total > i:
                payload = {
                    "query": {
                        "type": "group",
                        'logical_operator': 'and',
                        'nodes': [
                            {
                                "type": "terminal",
                                "service": "text",
                                "parameters": {"operator": "exact_match", "value": "Protein (only)", "attribute": "rcsb_entry_info.selected_polymer_entity_types"}
                            },
                            {
                                "type": "terminal",
                                "service": "text",
                                "parameters": {"attribute": "rcsb_entry_info.deposited_polymer_entity_instance_count", "operator": "equals", "value": 1}
                            },
                            *[
                                {
                                    "type": "terminal",
                                    "service": "text",
                                    "parameters": {k:v for k,v in zip(['attribute','operator','value'], q)}
                                }
                            for q in self.query],
                        ],
                    },
                    "request_options": {
                        "group_by": {
                            "aggregation_method": "sequence_identity",
                            "similarity_cutoff": 100,
                        },
                        "group_by_return_type": "representatives",
                        "paginate": {"start": i, "rows": i+batch_size}
                    },
                    "return_type": "polymer_entity"
                }
                r = requests.get(f'https://search.rcsb.org/rcsbsearch/v2/query?json={json.dumps(payload)}')
                try:
                    response_dict = json.loads(r.text)
                    ids.extend([x['identifier'].split('_')[0] for x in response_dict['result_set']])
                except:
                    print('An error occured when querying RCSB.')
                    print(r.text)
                    exit()
                if total is None:
                    total = response_dict['group_by_count']
                i += batch_size

        ids = sorted(list(set(ids))) # filter identical ids
        random.seed(42)
        random.shuffle(ids) # for reproducible subsampling when using self.limit
        ids = ids[:self.limit] # for testing

        n_jobs = min(self.n_jobs, self.max_requests) # RCSB has a request limit
        if n_jobs < 1:
            n_jobs = self.max_requests

        failed = Parallel(n_jobs=n_jobs)(delayed(self.download_from_rcsb)(id) for id in tqdm(ids, desc='Downloading PDBs'))
        failed = [f for f in failed if not f is True]
        if len(failed)>0:
            print(f'Failed to download {len(failed)} PDB files.')


    def download_from_rcsb(self, id):
        try:
            r = requests.get(f'https://data.rcsb.org/rest/v1/core/polymer_entity/{id}/1')
            obj = json.loads(r.text)
            download_url(f'https://files.rcsb.org/download/{id}.pdb.gz', f'{self.root}/raw/files', log=False)
            unzip_file(f'{self.root}/raw/files/{id}.pdb.gz')
            with open(f'{self.root}/raw/files/{id}.annot.json', 'w') as file:
                json.dump(obj, file)
            return True
        except KeyboardInterrupt:
            exit()
        except Exception as e:
            if os.path.exists(f'{self.root}/raw/files/{id}.pdb.gz'):
                os.remove(f'{self.root}/raw/files/{id}.pdb.gz')
            return id



'''
Tests all downloads with 'use_precomputed=False'. The number of downloaded files and the number of parsed files is patched to a small number where possible. However, most datasets require downloading one large file, which takes time. Hence removed from GitHub testing CI (by not naming it according to pytest convention).
'''

import unittest, tempfile, os, shutil, glob
from collections import defaultdict
from unittest import mock
from proteinshake.datasets import *

def download_mock(self):
    mock_data_path = os.path.dirname(os.path.realpath(__file__)) + '/mock_data'
    shutil.copytree(mock_data_path, f'{self.root}/raw/files', dirs_exist_ok=True)

def get_raw_files_mock(self):
    return glob.glob(f'{self.root}/raw/files/????.pdb')[:self.limit]

def get_id_from_filename_mock(self, filename):
    return filename[:4]

def scop_download_mock(self):
    download_mock(self)
    self.scop = self._parse_scop(f'{self.root}/raw/files/scop.txt')

def pli_download_mock(self):
    download_mock(self)
    self.index_data = self.parse_pdbbind_PL_index(f'{self.root}/raw/files/INDEX_refined_data.{self.version}')

def GODag_mock(*args, **kwargs):
    class DummyTerm:
        def __init__(self):
            self.namespace = 'molecular_function'
    factory = lambda: DummyTerm()
    return defaultdict(factory)

class TestDownload(unittest.TestCase):

    @mock.patch.object(ProteinLigandInterfaceDataset, 'download', pli_download_mock)
    @mock.patch.object(ProteinLigandInterfaceDataset, 'get_raw_files', get_raw_files_mock)
    def test_protein_ligand(self):
        with tempfile.TemporaryDirectory() as tmp:
            ds = ProteinLigandInterfaceDataset(root=tmp, use_precomputed=False)
    
    @mock.patch.object(ProteinProteinInterfaceDataset, 'download', download_mock)
    @mock.patch.object(ProteinProteinInterfaceDataset, 'get_raw_files', get_raw_files_mock)
    def test_protein_protein(self):
        with tempfile.TemporaryDirectory() as tmp:
            ds = ProteinProteinInterfaceDataset(root=tmp, use_precomputed=False)

    @mock.patch('proteinshake.datasets.RCSBDataset.limit', return_value=5)
    #@mock.patch.object(RCSBDataset, 'download', download_mock)
    #@mock.patch.object(RCSBDataset, 'get_raw_files', get_raw_files_mock)
    def test_rcsb(self, mock):
        with tempfile.TemporaryDirectory() as tmp:
            ds = RCSBDataset(root=tmp, use_precomputed=False)

    @mock.patch.object(GeneOntologyDataset, 'download', download_mock)
    @mock.patch.object(GeneOntologyDataset, 'get_raw_files', get_raw_files_mock)
    @mock.patch('proteinshake.datasets.gene_ontology.GODag', GODag_mock)
    def test_go(self):
        with tempfile.TemporaryDirectory() as tmp:
            ds = GeneOntologyDataset(root=tmp, use_precomputed=False)

    @mock.patch.object(EnzymeCommissionDataset, 'download', download_mock)
    @mock.patch.object(EnzymeCommissionDataset, 'get_raw_files', get_raw_files_mock)
    def test_ec(self):
        with tempfile.TemporaryDirectory() as tmp:
            ds = EnzymeCommissionDataset(root=tmp, use_precomputed=False)

    @mock.patch.object(ProteinFamilyDataset, 'download', download_mock)
    @mock.patch.object(ProteinFamilyDataset, 'get_raw_files', get_raw_files_mock)
    def test_pfam(self):
        with tempfile.TemporaryDirectory() as tmp:
            ds = ProteinFamilyDataset(root=tmp, use_precomputed=False)

    @mock.patch.object(TMAlignDataset, 'download', download_mock)
    @mock.patch.object(TMAlignDataset, 'get_raw_files', get_raw_files_mock)
    def test_tm(self):
        with tempfile.TemporaryDirectory() as tmp:
            ds = TMAlignDataset(root=tmp, use_precomputed=False)

    @mock.patch('proteinshake.datasets.AlphaFoldDataset.limit', return_value=5)
    #@mock.patch.object(AlphaFoldDataset, 'download', download_mock)
    #@mock.patch.object(AlphaFoldDataset, 'get_raw_files', get_raw_files_mock)
    #@mock.patch.object(AlphaFoldDataset, 'get_id_from_filename', get_id_from_filename_mock)
    def test_af(self, mock):
        with tempfile.TemporaryDirectory() as tmp:
            organism = 'methanocaldococcus jannaschii'
            ds = AlphaFoldDataset(root=tmp, organism=organism, use_precomputed=False)

    @mock.patch.object(SCOPDataset, 'download', scop_download_mock)
    @mock.patch.object(SCOPDataset, 'get_raw_files', get_raw_files_mock)
    def test_scop(self):
        with tempfile.TemporaryDirectory() as tmp:
            ds = SCOPDataset(root=tmp, use_precomputed=False)

    @mock.patch.object(ProteinLigandDecoysDataset, 'download', download_mock)
    @mock.patch.object(ProteinLigandDecoysDataset, 'get_raw_files', get_raw_files_mock)
    def test_decoys(self):
        with tempfile.TemporaryDirectory() as tmp:
            ds = ProteinLigandDecoysDataset(root=tmp, use_precomputed=False)

if __name__ == '__main__':
    unittest.main()



import itertools

from scipy.stats import spearmanr
import numpy as np
from sklearn import metrics
from sklearn.model_selection import train_test_split

from proteinshake.datasets import TMAlignDataset
from proteinshake.tasks import Task

class StructureSimilarityTask(Task):
    """ Predict the structural similarity between two proteins. This is a pair-wise protein-level regression task.
    Ground truth is computed using the TMAlign software. Split indices are stored as tuples which contain two indices in
    the underlying dataset.
    """

    DatasetClass = TMAlignDataset

    def __init__(self, *args, **kwargs):

        super().__init__(*args, **kwargs)

    def update_index(self):
        """ Transform to pairwise indexing """
        self.train_index = self.compute_pairs(self.train_index)
        self.val_index = self.compute_pairs(self.val_index)
        self.test_index = self.compute_pairs(self.test_index)
        
    def compute_targets(self):
        self.train_targets = np.array([self.target(*self.proteins[i]) for i in self.train_index])
        self.val_targets = np.array([self.target(*self.proteins[i]) for i in self.val_index])
        self.test_targets = np.array([self.target(*self.proteins[i]) for i in self.test_index])

    @property
    def task_type(self):
        return ('protein_pair', 'regression')

    def compute_pairs(self, index):
        combinations = np.array(list(itertools.combinations(range(len(index)), 2)), dtype=int)
        return index[combinations]

    def target(self, protein1, protein2):
        pdbid_1 = protein1['protein']['ID']
        pdbid_2 = protein2['protein']['ID']
        return self.dataset.lddt(pdbid_1,pdbid_2)

    def dummy_output(self):
        import random
        return [random.random() for _ in range(len(self.test_targets))]

    def evaluate(self, y_pred):
        return {
            'mse': metrics.mean_squared_error(self.test_targets, y_pred),
            'r2':  spearmanr(self.test_targets, y_pred)
        }



'''
Tests prediction task classes.
'''

import random
import unittest, tempfile
from proteinshake.tasks import *


class TestTasks(unittest.TestCase):

    def task_check(self, task, skip_splits=False, pair=False):
        # check splits
        if not skip_splits:
            assert len(task.train_index) > 0
            assert len(task.test_index) > 0
            assert len(task.val_index) > 0
            assert len(set(task.train_index.flatten()).intersection(set(task.test_index.flatten()))) == 0
            assert len(set(task.train_index.flatten()).intersection(set(task.val_index.flatten()))) == 0
            assert len(set(task.val_index.flatten()).intersection(set(task.test_index.flatten()))) == 0

        # check targets
        prots = task.dataset.proteins()
        protein = next(prots)
        if pair:
            protein_2 = next(task.dataset.proteins())
            self.assertIsNotNone(task.target(protein, protein_2))
        else:
            self.assertIsNotNone(task.target(protein))

        # check eval
        pred = task.dummy_output()
        self.assertIsNotNone(task.evaluate(pred))

    def test_enzyme_class(self):
        with tempfile.TemporaryDirectory() as tmp:
            self.task_check(EnzymeClassTask(split='random', root=tmp))
            self.task_check(EnzymeClassTask(split='sequence', root=tmp))
            self.task_check(EnzymeClassTask(split='structure', root=tmp))

    def test_protein_family(self):
        with tempfile.TemporaryDirectory() as tmp:
            self.task_check(ProteinFamilyTask(split='random', root=tmp))
            self.task_check(ProteinFamilyTask(split='sequence', root=tmp))
            self.task_check(ProteinFamilyTask(split='structure', root=tmp))

    def test_gene_ontology(self):
        with tempfile.TemporaryDirectory() as tmp:
            self.task_check(GeneOntologyTask(split='random', root=tmp))
            self.task_check(GeneOntologyTask(split='sequence', root=tmp))
            self.task_check(GeneOntologyTask(split='structure', root=tmp))

    def test_binding_site_detection(self):
        with tempfile.TemporaryDirectory() as tmp:
            self.task_check(BindingSiteDetectionTask(split='random', root=tmp))
            self.task_check(BindingSiteDetectionTask(split='sequence', root=tmp))
            self.task_check(BindingSiteDetectionTask(split='structure', root=tmp))

    def test_ligand_affinity(self):
        with tempfile.TemporaryDirectory() as tmp:
            self.task_check(LigandAffinityTask(split='random', root=tmp))
            self.task_check(LigandAffinityTask(split='sequence', root=tmp))
            self.task_check(LigandAffinityTask(split='structure', root=tmp))

    def test_protein_protein_interface(self):
        with tempfile.TemporaryDirectory() as tmp:
            self.task_check(ProteinProteinInterfaceTask(split='random', root=tmp))
            self.task_check(ProteinProteinInterfaceTask(split='sequence', root=tmp))
            self.task_check(ProteinProteinInterfaceTask(split='structure', root=tmp))

    def test_structural_class(self):
        with tempfile.TemporaryDirectory() as tmp:
            self.task_check(StructuralClassTask(split='random', root=tmp))
            self.task_check(StructuralClassTask(split='sequence', root=tmp))
            self.task_check(StructuralClassTask(split='structure', root=tmp))

    def test_structure_search(self):
        with tempfile.TemporaryDirectory() as tmp:
            self.task_check(StructureSearchTask(split='random', root=tmp))
            self.task_check(StructureSearchTask(split='sequence', root=tmp))
            self.task_check(StructureSearchTask(split='structure', root=tmp))

    def test_virtual_screen(self):
        with tempfile.TemporaryDirectory() as tmp:
            self.task_check(VirtualScreenTask(root=tmp), skip_splits=True)

    def test_structure_similarity(self):
        with tempfile.TemporaryDirectory() as tmp:
            self.task_check(StructureSimilarityTask(split='random', root=tmp), pair=True)
            self.task_check(StructureSimilarityTask(split='sequence', root=tmp), pair=True)
            self.task_check(StructureSimilarityTask(split='structure', root=tmp), pair=True)

if __name__ == '__main__':
    unittest.main()



import os
import json
import itertools

import numpy as np
from sklearn.model_selection import train_test_split

from proteinshake.utils import download_url, save, load

class Task:
    """ Base class for task-related utilities.
    This class wraps a proteinshake dataset and exposes split indices,
    integer-coded labels for classification tasks, and an evaluator function.

    Sample usage (assuming you have a model in the namespace):

     .. code-block:: python

        >>> from proteinshake.tasks import EnzymeClassTask
        >>> task = EnzymeClassTask()
        >>> data = task.dataset.to_graph(eps=8).pyg()
        >>> y_pred = model(data[task.train])
        >>> task.evaluate(y_pred)
        ... {'roc_auc_score': 0.7}


    Arguments
    ----------
    dataset: pytorch.datasets.Dataset
        Dataset to use for this task.
    split: str, default='random'
        How to split the data. Can be 'random', 'sequence', or 'structure'.
    split_similarity_threshold: float
        Maximum similarity to allow between train and test samples.
    use_precomputed: bool, default=True
    """

    DatasetClass = None

    def __init__(self,
                 root                       = 'data',
                 split                      = 'random',
                 split_similarity_threshold = 0.7,
                 use_precomputed            = True,
                 **kwargs
                ):
        self.root = root
        self.dataset = self.DatasetClass(root=root, use_precomputed=use_precomputed)
        proteins = self.dataset.proteins()
        self.size = len(proteins)
        self.split_similarity_threshold = split_similarity_threshold
        self.split = split
    
        class Proteins(): # dummy class to implement __getitem__, could be implemented directly on the task
            def __init__(self, proteins):
                self.proteins = list(proteins)

            def __len__(self):
                return len(self.proteins)

            def __getitem__(self, idx):
                try:
                    idx = int(idx)
                except:
                    return [self.__getitem__(i) for i in idx]
                if idx >= len(self.proteins):
                    raise StopIteration
                return self.proteins[idx]

        self.proteins = Proteins(proteins)
        self.name = self.__class__.__name__

        # load split indices
        if not self.split == 'none':
            self.compute_index()
            self.compute_targets()

    def compute_index(self):
        split_name = f'{self.split}_split_{self.split_similarity_threshold}' if self.split in ['sequence','structure'] else f'{self.split}_split'
        if split_name in self.proteins[0]['protein']:
            self.train_index = np.array([i for i,p in enumerate(self.proteins) if p['protein'][split_name] == 'train'])
            self.val_index = np.array([i for i,p in enumerate(self.proteins) if p['protein'][split_name] == 'val'])
            self.test_index = np.array([i for i,p in enumerate(self.proteins) if p['protein'][split_name] == 'test'])
        else:
            self.train_index, self.val_index, self.test_index = self.compute_custom_split(self.split)

        self.update_index()

    def update_index(self):
        pass

    def compute_targets(self):
        # compute targets (e.g. for scaling)
        self.train_targets = np.array([self.target(self.proteins[i]) for i in self.train_index], dtype=object)
        self.val_targets = np.array([self.target(self.proteins[i]) for i in self.val_index], dtype=object)
        self.test_targets = np.array([self.target(self.proteins[i]) for i in self.test_index], dtype=object)
            
    def compute_custom_split(self, split):
        """ Implements custom splitting. Only necessary when not using the precomputed splits, e.g. when implementing a custom task.
        Note that the random, sequence and structure splits will be automatically computed for your custom task if it is merged into ProteinShake main.
        Compare also the proteinshake_release repository.

        Arguments
        ------------
        split: str
            Name of the custom split as passed to the task. ('random', 'sequence', 'structure', 'none')

        Returns:
        --------
        train_index
            Numpy array with the index of proteins in the train split.
        val_index
            Numpy array with the index of proteins in the validation split.
        test_index
            Numpy array with the index of proteins in the test split.
        """
        raise Exception('The requested split is not available. Implement <compute_custom_split> to provide your own splitting logic.')

    @property
    def task_type(self):
        """ Returns a string describing the type of task."""
        raise NotImplementedError

    @property
    def num_features(self):
        """ Number of input features to use for this task """
        raise NotImplementedError

    @property
    def num_classes(self):
        """ Size of the output dimension for this task """
        raise NotImplementedError

    @property
    def target(self, protein):
        """ Return the prediction target for one protein in the dataset.

        Arguments
        ------------
        protein: dict
            proteinshake protein dictionary


        .. code-block: python

            >>> from proteinshake.tasks import EnzymeCommissionTask
            >>> ta = EnzymeCommissionTask()
        """
        raise NotImplementedError

    def evaluate(self, y_pred):
        """ Evaluates prediction quality.

        Arguments
        -----------
        y_pred: list
            List of predicted outputs.

        Returns:
        --------
        dict
            Dictionary with evaluation results. Key-value pairs correspond to metric-score pairs. E.g. 'roc-auc': 0.7
        """
        raise NotImplementedError

    @property
    def train(self):
        return self.dataset[self.train_index]

    @property
    def val(self):
        return self.dataset[self.val_index]

    @property
    def test(self):
        return self.dataset[self.test_index]

    def to_graph(self, *args, **kwargs):
        self.dataset = self.dataset.to_graph(*args, **kwargs)
        return self

    def to_point(self, *args, **kwargs):
        self.dataset = self.dataset.to_point(*args, **kwargs)
        return self

    def to_voxel(self, *args, **kwargs):
        self.dataset = self.dataset.to_voxel(*args, **kwargs)
        return self

    def pyg(self, *args, **kwargs):
        self.dataset = self.dataset.pyg(*args, **kwargs)
        return self

    def dgl(self, *args, **kwargs):
        self.dataset = self.dataset.dgl(*args, **kwargs)
        return self

    def nx(self, *args, **kwargs):
        self.dataset = self.dataset.nx(*args, **kwargs)
        return self

    def np(self, *args, **kwargs):
        self.dataset = self.dataset.np(*args, **kwargs)
        return self

    def tf(self, *args, **kwargs):
        self.dataset = self.dataset.tf(*args, **kwargs)
        return self

    def torch(self, *args, **kwargs):
        self.dataset = self.dataset.torch(*args, **kwargs)
        return self



import os
from tqdm import tqdm
from proteinshake.utils import save, load, fx2str

class FrameworkDataset():
    """ Dataset base class for different frameworks.

    Parameters
    ----------
    data_list: generator
        A generator of objects from a representation.
    size: int
        The size of the dataset.
    path: str
        Path to save the processed dataset.
    transform: function
        A transform function to be applied in the __getitem__ method. Signature: transform(data, protein_dict) -> (data, protein_dict)
    pre_transform: function
        A transform function to be applied before writing the data. Signature: transform((data, protein_dict)) -> (data, protein_dict)
    pre_filter: function
        A filter function to be applied before writing the data. Signature: transform(data, protein_dict) -> bool
    """

    def __init__(self, data_list, size, path, transform=None, pre_transform=None, pre_filter=None):
        os.makedirs(path, exist_ok=True)
        self.path = path
        self.transform = transform
        self.pre_transform = pre_transform
        self.pre_filter = pre_filter
        transforms_repr = fx2str(pre_transform) + fx2str(pre_filter)
        if not os.path.exists(f'{path}/{size-1}.pkl'):
            i = 0
            for data_item in tqdm(data_list, desc='Converting', total=size):
                data = self.convert_to_framework(data_item)
                protein_dict = data_item.protein_dict
                if not self.pre_filter is None and not self.pre_filter(data, protein_dict):
                    continue
                if not self.pre_transform is None:
                    data, protein_dict = self.pre_transform(data, protein_dict)
                save((data, protein_dict), f'{path}/{i}.pkl')
                i += 1
            save(i,f'{path}/size.pkl')
            save(transforms_repr,f'{path}/transforms.pkl')
        self.size = load(f'{path}/size.pkl')
        original_repr = load(f'{path}/transforms.pkl')
        assert original_repr == transforms_repr, f'The pre_transform and/or pre_filter are not the same as when the dataset was created. If you want to change them, delete the folder at {path}'

    def convert_to_framework(self, data_item):
        return data_item.data

    def load_transform(self, data, protein_dict):
        return data, protein_dict

    def __len__(self):
        return self.size

    def __getitem__(self, idx):
        try:
            idx = int(idx)
        except:
            return [self.__getitem__(i) for i in idx]
        if idx > self.size - 1:
            raise StopIteration
        data, protein_dict = self.load_transform(*load(f'{self.path}/{idx}.pkl'))
        if not self.transform is None:
            return self.transform((data, protein_dict))
        else:
            return data, protein_dict

    def len(self):
        pass

    def get(self):
        pass



import numpy as np

from proteinshake.transforms import Transform


def _get_coords_array(protein, resolution='residue'):
    """ Get a numpy array of the protein coordinates


    Arguments
    ---------
    protein: dict
        A protein dictionary.

    resolution: str
        The resolution at which to return the coordinates ('residue', or 'atom')

    Returns
    --------
    np.array
        Numpy array of shape (n_(residues or atoms),  3)

    """

    N = len(protein[resolution]['x'])

    return np.array([protein[resolution]['x'],
                     protein[resolution]['y'],
                     protein[resolution]['z']
                     ]
                   ).T.reshape((N, 3, 1))

def _set_coords(protein, coord_array, resolution='residue', in_place=True):
    """ Given an Nx3 array of coordinates, set them to the
    protein dictionary coord key.

    Arguments
    ----------
    protein: dict
        Proteinshake protein dict
    resolution: str
        Which resolution to use ('residue' or 'atom')
    coord_array: np.array
        Nx3 numpy coordinate array

    """

    protein[resolution]['x'] = list(coord_array[:,0])
    protein[resolution]['y'] = list(coord_array[:,1])
    protein[resolution]['z'] = list(coord_array[:,2])

class CenterTransform(Transform):
    """ Center the coordinates of a protein at atom and residue level.
    We use the Ca to compute the center for all the atoms.
        """
    def __init__(self, resolution='residue'):
        self.resolution = resolution
        super().__init__()

    def __call__(self, protein):
        coords = _get_coords_array(protein, resolution=self.resolution)

        center = np.mean(coords, axis=0)
        new_coords = coords - center

        _set_coords(protein, new_coords, resolution=self.resolution)

        return protein

class RandomRotateTransform(Transform):
    """ Apply a random rotation to the coordinate arrays of a protein"""

    def __init__(self, resolution='residue', seed=42):
        self.seed = seed
        self.resolution = resolution
        np.random.seed(self.seed)
        super().__init__()
        pass

    def __call__(self, protein):
        coords = _get_coords_array(protein, resolution=self.resolution)

        # choose an angle and axis along which to rotate
        rnd = int(np.random.randint(0,2+1,(1,)))
        rot = int(np.random.randint(1,3+1,(1,)))
        rotation_plane = {0:[0,1],1:[1,2],2:[0,2]}[rnd]
        N = len(protein[self.resolution]['x'])
        coords_rot = np.rot90(coords,k=rot, axes=rotation_plane).reshape(N, 3)

        _set_coords(protein, coords_rot, resolution=self.resolution)

        return protein



import numpy as np
from sklearn import metrics

from proteinshake.datasets import ProteinProteinInterfaceDataset
from proteinshake.tasks import Task
from proteinshake.transforms import CenterTransform, RandomRotateTransform, Compose

class ProteinProteinInterfaceTask(Task):
    """ Identify the binding residues of a protein-protein complex. This is a residue-level binary classification.

    NOTE: To make this an interesting task, the loader has to
    split the protein into its chains so that the model only sees
    one chain at a time. You should also pass the following transforms to the dataset
    :meth:`proteinshake.transforms.CenterTransform` and :meth:`proteinshake.transforms.RandomRotateTransform` when using representations that are aware of the atomic coordinates.

    NOTE: This task is currently in beta.


    .. code-block:: python

        >>> from proteinshake.tasks import ProteinProteinInterfaceTask
        >>> from proteinshake.transforms import CenterTransform, RandomRotateTransform
        >>> ta = ProteinProteinInterfaceTask()
        >>> data = ta.dataset.to_voxel(transforms=[CenterTransform(), RandomRotateTransform()).torch()
    """

    DatasetClass = ProteinProteinInterfaceDataset

    @property
    def task_type(self):
        return ('residue', 'binary') 

    def dummy_output(self):
        import random
        return [random.randint(0, 1) for p in self.test_targets]

    def compute_targets(self):
        # compute targets (e.g. for scaling)
        self.train_targets = [p for i in self.train_index for p in self.target(self.proteins[i])]
        self.val_targets = [p for i in self.val_index for p in self.target(self.proteins[i])]
        self.test_targets = [p for i in self.test_index for p in self.target(self.proteins[i])]

    def target(self, protein):
        return protein['residue']['is_interface']

    def evaluate(self, y_pred):
        """ Evaluate performance of an interface classifier.

        y_pred: list
            One list for each protein in the test set with a probability for each residue in the protein.

        """
        labels = [self.target(p) for p in \
                  self.proteins[self.test_index]]
        labels = np.hstack(labels)
        y_pred = np.hstack(y_pred)
        return {
                'auc-roc': metrics.roc_auc_score(labels, y_pred),
                'average precision': metrics.average_precision_score(labels, y_pred),
                }

    def to_graph(self, *args, **kwargs):
        self.dataset = self.dataset.to_graph(*args, **kwargs, transform=Compose([CenterTransform(), RandomRotateTransform()]))
        return self

    def to_point(self, *args, **kwargs):
        self.dataset = self.dataset.to_point(*args, **kwargs, transform=Compose([CenterTransform(), RandomRotateTransform()]))
        return self

    def to_voxel(self, *args, **kwargs):
        self.dataset = self.dataset.to_voxel(*args, **kwargs, transform=Compose([CenterTransform(), RandomRotateTransform()]))
        return self



from sklearn import metrics
from functools import cached_property

from proteinshake.datasets import SCOPDataset
from proteinshake.tasks import Task

class StructuralClassTask(Task):
    """ Predict the SCOP class of a protein structure. This is a protein-level multi-class prediction.

    """

    DatasetClass = SCOPDataset
    
    def __init__(self, scop_level='SCOP-FA', *args, **kwargs):
        self.scop_level = scop_level
        super().__init__(*args, **kwargs)

    @cached_property
    def token_map(self):
        labels = {p['protein'][self.scop_level] for p in self.proteins}
        return {label: i for i, label in enumerate(sorted(list(labels)))}

    @property
    def num_classes(self):
        return len(self.token_map)

    def dummy_output(self):
        import random
        tokens = list(self.token_map.values())
        return [random.choice(tokens) for _ in range(len(self.test_targets))]

    @property
    def task_type(self):
        return ('protein', 'multi-class')

    @property
    def num_features(self):
        return 20

    def target(self, protein):
        return self.token_map[protein['protein'][self.scop_level]]

    def evaluate(self, y_pred):
        """ Using metrics from https://doi.org/10.1073/pnas.1821905116 """
        return {
            'precision': metrics.precision_score(list(self.test_targets), y_pred, average='macro', zero_division=0),
            'recall': metrics.recall_score(list(self.test_targets), y_pred, average='macro', zero_division=0),
            'accuracy': metrics.accuracy_score(list(self.test_targets), y_pred),
            #'AUROC': metrics.roc_auc_score(self.test_targets, y_pred, average='macro', multi_class='ovo'),
        }



import os
from tqdm import tqdm
import numpy as np

from proteinshake.utils import tokenize

class Point():
    """ Point representation of a protein.

    Parameters
    ----------
    protein: dict
        A protein object.

    """

    def __init__(self, protein):
        resolution = 'atom' if 'atom' in protein else 'residue'
        self.protein_dict = protein
        self.resolution = resolution
        labels = tokenize(protein[resolution][f'{resolution}_type'], resolution=resolution)
        coords =  np.stack([protein[resolution]['x'], protein[resolution]['y'], protein[resolution]['z']], axis=1)
        self.data = np.concatenate((coords,np.expand_dims(labels,-1)), axis=-1)



class PointDataset():
    """ Point representation of a protein structure dataset.

    Parameters
    ----------
    proteins: generator
        A generator of protein objects from a Dataset.
    size: int
        The size of the dataset.
    path: str
        Path to save the processed dataset.
    resolution: str, default 'residue'
        Resolution of the proteins to use in the graph representation. Can be 'atom' or 'residue'.

    """

    def __init__(self, proteins, root, name, resolution='residue'):
        self.path = f'{root}/processed/point/{name}_{resolution}'
        self.points = (Point(protein) for protein in proteins)
        self.size = len(proteins)

    def torch(self, *args, **kwargs):
        from proteinshake.frameworks.torch import TorchPointDataset
        return TorchPointDataset(self.points, self.size, self.path+'.torch', *args, **kwargs)

    def tf(self, *args, **kwargs):
        from proteinshake.frameworks.tf import TensorflowPointDataset
        return TensorflowPointDataset(self.points, self.size, self.path+'.tf', *args, **kwargs)

    def np(self, *args, **kwargs):
        from proteinshake.frameworks.np import NumpyPointDataset
        return NumpyPointDataset(self.points, self.size, self.path+".np", *args, **kwargs)



import tensorflow as tf
from proteinshake.frameworks.dataset import FrameworkDataset

class TensorflowVoxelDataset(FrameworkDataset):

    def convert_to_framework(self, data_item):
        return tf.sparse.from_dense(tf.convert_to_tensor(data_item.data, dtype=tf.float32))

    def load_transform(self, data, protein_dict):
        return tf.sparse.to_dense(data), protein_dict


class TensorflowPointDataset(FrameworkDataset):

    def convert_to_framework(self, data_item):
        return tf.convert_to_tensor(data_item.data, dtype=tf.float32)



"""
Some basic functions for embedding protein sequences. Supply to a representation class.

All embeddings take a sequence string as an input and return the embedding as a numpy array of shape n x d.
"""

import numpy as np

residue_alphabet = 'ARNDCEQGHILKMFPSTWYV'
atom_alphabet = 'NCOSH'

def onehot(sequence, resolution='residue'):
    """ Compute the one-hot encoding of a protein sequence.

    Parameters
    ----------
    sequence: str
        The protein sequence.
    resolution: str, default 'resolution'
        Resolution of the protein. 'residue' or 'atom'.

    Returns
    -------
    ndarray
        The embedded sequence.
    """
    if resolution == 'residue':
        return np.stack([np.eye(len(residue_alphabet))[residue_alphabet.index(aa)] for aa in sequence])
    else:
        return np.stack([np.eye(len(atom_alphabet))[atom_alphabet.index(aa[0])] for aa in sequence])



def tokenize(sequence, resolution='residue'):
    """ Tokenizes the sequence.

    Parameters
    ----------
    sequence: str
        The protein sequence.
    resolution: str, default 'resolution'
        Resolution of the protein. 'residue' or 'atom'.

    Returns
    -------
    ndarray
        The embedded sequence.
    """
    if resolution == 'residue':
        return np.array([residue_alphabet.index(aa) for aa in sequence])
    else:
        return np.array([atom_alphabet.index(aa[0]) for aa in sequence])

# from: https://gist.github.com/foowaa/5b20aebd1dff19ee024b6c72e14347bb
def sinusoid_encoding_table(n_position, d_hid, padding_idx=None):
    """ Helper function to build a sinusoidal lookup table for positional encodings.
    """
    def cal_angle(position, hid_idx):
        return position / np.power(10000, 2 * (hid_idx // 2) / d_hid)
    def get_posi_angle_vec(position):
        return [cal_angle(position, hid_j) for hid_j in range(d_hid)]
    sinusoid_table = np.array([get_posi_angle_vec(pos_i) for pos_i in range(n_position)])
    sinusoid_table[:, 0::2] = np.sin(sinusoid_table[:, 0::2])
    sinusoid_table[:, 1::2] = np.cos(sinusoid_table[:, 1::2])
    if padding_idx is not None:
        sinusoid_table[padding_idx] = 0.
    return sinusoid_table

def positional_encoding(sequence, dim=128):
    """ Sinusoidal encoding of sequence position.

    Parameters
    ----------
    sequence: str
        The protein sequence

    Returns
    -------
    ndarray
        The embedded sequence.
    """
    n = len(sequence)
    table = sinusoid_encoding_table(n, dim)
    return table

def compose_embeddings(embeddings):
    """ Composes multiple embeddings into one by concatenating the results.

    Parameters
    ----------
    embeddings: list
        A list of embeddings

    Returns
    -------
    function
        A substitute embedding function.
    """

    return lambda sequence: np.concatenate([e(sequence) for e in embeddings], axis=-1)



from .transforms import *
from .coords import *

__all__ = [
            'Compose',
            'Transform',
            'IdentityTransform',
            'CenterTransform',
            'RandomRotateTransform',
          ]

classes = __all__



import json, os
from goatools.obo_parser import GODag

from proteinshake.datasets import RCSBDataset
from proteinshake.utils import download_url, unzip_file
from functools import cached_property

class GeneOntologyDataset(RCSBDataset):
    """ Proteins from RCSB for which the Gene Ontology (GO) term is known.
    Each protein in the dataset has a `GO` attribute which stores the path
    from the root to the leaves along the GO hierarchy.

    .. list-table:: Dataset stats
       :widths: 100
       :header-rows: 1

       * - # proteins
       * - 32633


   .. list-table:: Annotations
      :widths: 25 25 50
      :header-rows: 1

      * - Attribute
        - Key
        - Sample value
      * - Molecular Function
        - :code:`protein['protein']['molecular_function']`
        - :code:`['GO:0003674', 'GO:0005198']`
      * - Localization
        - :code:`protein['protein']['cellular_component']`
        - :code:`['GO:0005575', 'GO:0018995',..]`
      * - Biological process
        - :code:`protein['protein']['biological_process']`
        - ...


    """

    additional_files = ['GeneOntologyDataset.godag.obo']

    def __init__(self, query=[['rcsb_polymer_entity_annotation.type','exact_match','GO']], **kwargs):
        super().__init__(query=query, **kwargs)

    @cached_property
    def godag(self):
        if not os.path.exists(f'{self.root}/{self.name}.godag.obo'):
            download_url(f'{self.repository_url}/{self.name}.godag.obo.gz', f'{self.root}')
            unzip_file(f'{self.root}/{self.name}.godag.obo.gz')
        return GODag(f'{self.root}/{self.name}.godag.obo', prt=None)

    def download(self):
        super().download()
        if not os.path.exists(f'{self.root}/{self.name}.godag.obo'):
            download_url(f'http://current.geneontology.org/ontology/go-basic.obo', f'{self.root}', log=False)
            os.rename(f'{self.root}/go-basic.obo', f'{self.root}/{self.name}.godag.obo')

    def add_protein_attributes(self, protein):
        godag = GODag(f'{self.root}/{self.name}.godag.obo', prt=None) # cannot use self.godag because the GODAG is not pickleable (for the release)
        with open(f'{self.root}/raw/files/{protein["protein"]["ID"]}.annot.json','r') as file:
            annot = json.load(file)
        go_terms = []
        if not 'rcsb_polymer_entity_annotation' in annot: return None
        for a in annot['rcsb_polymer_entity_annotation']:
            if a['type'] == 'GO':
                go_terms.extend([go['id'] for go in a['annotation_lineage']])
        protein['protein']['molecular_function'] = [term for term in go_terms if godag[term].namespace == 'molecular_function']
        protein['protein']['cellular_component'] = [term for term in go_terms if godag[term].namespace == 'cellular_component']
        protein['protein']['biological_process'] = [term for term in go_terms if godag[term].namespace == 'biological_process']
        return protein

    def describe(self):
        desc = super().describe()
        desc['property'] = "Gene Ontology (GO)"
        desc['values'] = f"{len(set((p['GO'][0] for p in self.proteins)))} (root)"
        desc['type'] = 'Categorical, Hierarchical'
        return desc



import json

from proteinshake.datasets import RCSBDataset

class ProteinFamilyDataset(RCSBDataset):
    """ Proteins from RCSB for which the protein family (Pfam) term is known.
    Each protein in the dataset has a `Pfam` attribute which stores the list of protein families.

    .. list-table:: Dataset stats
       :widths: 100
       :header-rows: 1

       * - # proteins
       * - 31109


   .. list-table:: Annotations
      :widths: 25 35 45
      :header-rows: 1

      * - Attribute
        - Key
        - Sample value
      * - Pfam accession code
        - :code:`protein['protein']['Pfam']`
        - `['PF00102']`

    """

    def __init__(self, pfam_version='34.0', query=[['rcsb_polymer_entity_annotation.type','exact_match','Pfam']], **kwargs):
        self.pfam_version = pfam_version
        super().__init__(query=query, **kwargs)

    def add_protein_attributes(self, protein):
        with open(f'{self.root}/raw/files/{protein["protein"]["ID"]}.annot.json','r') as file:
            annot = json.load(file)
        pfams = []
        for a in annot['rcsb_polymer_entity_annotation']:
            if a['type'] == 'Pfam':
                # pfams.append(a['name'])
                pfams.append(a['annotation_id'])
        protein['protein']['Pfam'] = pfams
        return protein

    def describe(self):
        desc = super().describe()
        desc['property'] = "Protein Family (Pfam)"
        desc['values'] = f"{len(set((p['Pfam'][0] for p in self.proteins)))} (root)"
        desc['type'] = 'Categorical, Hierarchical'
        return desc



import itertools
from collections import defaultdict
from functools import cached_property

from scipy.stats import spearmanr
import numpy as np
from sklearn import metrics
from sklearn.model_selection import train_test_split

from proteinshake.datasets import TMAlignDataset
from proteinshake.tasks import Task

class StructureSearchTask(Task):
    """ Retrieve similar proteins to a query.
    Evaluation is cast in the setting of recommender systems where we wish
    to retrieve 'relevant' documents from a large pool of documents.
    Here, a protein is a document and the relevant ones are all proteins
    with a minimum similarity to the query protein.

    """

    DatasetClass = TMAlignDataset

    def __init__(self, min_sim=0.8, *args, **kwargs):
        self.min_sim = min_sim
        super().__init__(*args, **kwargs)

    @property
    def task_type(self):
        return ('protein', 'retrieval')

    @cached_property
    def targets(self):
        """ Precompute the set of similar proteins for each query """
        targets = {}
        for query in self.proteins:
            targets[query['protein']['ID']] = [c['protein']['ID'] for c in self.proteins if self.dataset.lddt(query['protein']['ID'], c['protein']['ID']) >= self.min_sim]
        return targets

    def target(self, protein):
        """ The target for a protein is a list of proteins deemed 'relevant'
        according to `self.min_sim`.
        """
        return self.targets[protein['protein']['ID']]

    def _precision_at_k(self, y_pred, targets, k):
        return len(set(y_pred[:k]).intersection(set(targets))) / len(y_pred)

    def _recall_at_k(self, y_pred, targets, k):
        return len(set(y_pred[:k]).intersection(set(targets))) / len(targets)

    def dummy_output(self):
        import random
        pred = []
        ids = [p['protein']['ID'] for p in self.proteins] 
        for query in self.proteins[self.test_index]:
            targets = self.target(query)
            pred.append(random.sample(ids, len(targets)))
        return pred

    def evaluate(self, y_pred, k=5):
        """ Retrieval metrics.

        Arguments
        -----------
        y_pred:
            List of indices of items (hits) in the dataset for a query.
        """
        results = defaultdict(list)
        for query, preds in zip(self.proteins[self.test_index], y_pred):
            targets = self.target(query)
            results['precision_at_k'].append(self._precision_at_k(preds, targets, k))
            results['recall_at_k'].append(self._recall_at_k(preds, targets, k))

        return {k: np.mean(v) for k, v in results.items()}



# -*- coding: utf-8 -*-
"""
Base dataset class for protein 3D structures.
"""
import os, gzip, inspect, time, itertools, tarfile, io
from collections import defaultdict
from functools import cached_property
import multiprocessing as mp

import pandas as pd
import numpy as np
import freesasa
from biopandas.pdb import PandasPdb
from tqdm import tqdm
from joblib import Parallel, delayed
from sklearn.neighbors import kneighbors_graph, radius_neighbors_graph
from fastavro import reader as avro_reader

from proteinshake.transforms import IdentityTransform
from proteinshake.utils import download_url, save, load, unzip_file, write_avro, Generator

AA_THREE_TO_ONE = {'ALA': 'A', 'CYS': 'C', 'ASP': 'D', 'GLU': 'E', 'PHE': 'F', 'GLY': 'G', 'HIS': 'H', 'ILE': 'I', 'LYS': 'K', 'LEU': 'L', 'MET': 'M', 'ASN': 'N', 'PRO': 'P', 'GLN': 'Q', 'ARG': 'R', 'SER': 'S', 'THR': 'T', 'VAL': 'V', 'TRP': 'W', 'TYR': 'Y'}
AA_ONE_TO_THREE = {v:k for k, v in AA_THREE_TO_ONE.items()}

# maps the date-format release to Zenodo identifier
RELEASES = {
    'latest': '1175739',
}

class Dataset():
    """ Base dataset class. Holds the logic for downloading and parsing PDB files.
    If ``use_precomputed=True``, fetched pre-processed data from Zenodo.
    Else, builds the dataset from scratch by executing: :meth:`download()` to fetch structures in PDB format, then :meth:`parse()` is applied to each to extract the relevant info and store it in a protein dictionary which has three outer keys ``'protein'``, ``'residue'``, and ``'atom'``. Subclassing :meth:`add_protein_attributes` lets the user include custom attributes.

    .. note::

        All child classes inherit these attributes and optionally add their own.

    .. list-table:: Annotations
      :widths: 25 35 45
      :header-rows: 1

      * - Attribute
        - Key
        - Sample value
      * - Protein identifier
        - :code:`protein['protein']['ID']`
        - ``'1JC8'``
      * - Sequence
        - :code:`protein['protein']['sequence']`
        - ``'IWGDSGKLITTA'``
      * - Assigned train/val/test split
        - :code:`protein['protein']['sequence_split_<CUTOFF>']`, :code:`protein['protein']['structure_split_<CUTOFF>']`
        - ``'train'``
      * - Residue position on chain
        - :code:`protein['residue']['residue_number']`
        - ``[2, 3, 4, 5, 6, 7, 8, 9, 10, ...,]``
      * - Amino acid type (single letter)
        - :code:`protein['residue']['residue_type']`
        - :code:`['I', 'W', 'G', 'D', 'S',..]`
      * - 3D coordinates
        - :code:`protein[{'residue' | 'atom'}][{'x'|'y'|'z'}]`
        - ``[5.191999912261963, 3.9860000610351562,..]``
      * - Solvent accessible surface area
        - :code:`protein[{'residue'|'atom'}]['SASA']`
        - :code:`[242.03138732910156, 136.46714782714844,... ]`
      * - Relative accessible surface area
        - :code:`protein['residue']['RSA']`
        - :code:`[1.377291202545166, 0.5476430058479309,... ]`
      * - Atom position
        - :code:`protein['atom']['atom_number']`
        - :code:`[7, 8, 9, 10, 11,..]`
      * - Atom type
        - :code:`protein['atom']['atom_type']`
        - :code:`['N', 'CA', 'C', 'O',...]`


    Arguments
    -----------
    root: str, default 'data'
        The data root directory to store both raw and parsed data.
    use_precomputed: bool, default True
        If `True`, will download the processed dataset from the ProteinShake repository (recommended). If `False`, will download the raw data from the original sources and process them on your device. You can use this option if you wish to create a custom dataset. Using `False` is compute-intensive, consider increasing `n_jobs`.
    release: str, default '12JUL2022'
        The tag of the dataset release. See https://github.com/BorgwardtLab/proteinshake/releases for all available releases. "latest" (default) is recommended.
    only_single_chain: bool, default False
        If `True`, will only use single-chain proteins.
    check_sequence: bool, default False
        If `True`, will discard proteins whose primary sequence is not identical with the sequence of amino acids in the structure. This can happen if the structure is not complete (e.g. for parts that could not be crystallized).
    n_jobs: int, default 1
        The number of jobs for downloading and parsing files. It is recommended to increase the number of jobs with `use_precomputed=False`.
    minimum_length: int, default 10
        Proteins smaller than minimum_length residues will be skipped.
    maximum_length: int, default 2048
        Proteins larger than maximum_length residues will be skipped.
    exclude_ids: list, default []
        Exclude PDB IDs from the dataset.
    """

    additional_files = [] # indicates the additional file names that are to be included in the release
    exlude_args_from_signature = []

    def __init__(self,
            root                           = 'data',
            use_precomputed                = True,
            release                        = 'latest',
            only_single_chain              = False,
            check_sequence                 = False,
            n_jobs                         = 1,
            minimum_length                 = 10,
            maximum_length                 = 2048,
            exclude_ids                    = [],
            ):
        self.repository_url = f'https://sandbox.zenodo.org/record/{RELEASES[release]}/files'
        self.n_jobs = n_jobs
        self.use_precomputed = use_precomputed
        self.root = root
        self.minimum_length = minimum_length
        self.maximum_length = maximum_length
        self.only_single_chain = only_single_chain
        self.check_sequence = check_sequence
        self.release = release
        self.exclude_ids = exclude_ids
        
        os.makedirs(f'{self.root}', exist_ok=True)
        self.check_signature()

        if not use_precomputed:
            self.start_download()
            self.parse()
        else:
            self.check_signature_same_as_hosted()

    def compute_signature(self, use_defaults=False):
        signature = dict(inspect.signature(self.__init__).parameters.items())
        class_object = self.__class__
        while True: # add base signatures to subclass signature
            signature = {**dict(inspect.signature(class_object.__init__).parameters.items()), **signature}
            if len(class_object.__bases__) == 0: break
            class_object = class_object.__bases__[0]
        arg_names = [n for n in signature.keys() if not n in ['self', 'args', 'kwargs', 'n_jobs', 'root']+self.exlude_args_from_signature]
        if use_defaults:
            return self.name + ' | ' + ', '.join([k + '=' + str(signature[k].default) for k in arg_names])
        return self.name + ' | ' + ', '.join([k + '=' + str(getattr(self, k)) for k in arg_names])

    @cached_property
    def default_signature(self):
        return self.compute_signature(use_defaults=True)

    @cached_property
    def signature(self):
        return self.compute_signature(use_defaults=False)

    def check_signature(self):
        if os.path.exists(f'{self.root}/signature.txt'):
            with open(f'{self.root}/signature.txt','r') as file:
                assert file.read() == self.signature, 'The Dataset is called with different arguments than were used to create it. Delete or change the root.'
        else:
            with open(f'{self.root}/signature.txt','w') as file:
                file.write(self.signature)

    def check_signature_same_as_hosted(self):
        """ Safety check to ensure the provided dataset arguments are the same as were used to precompute the datasets. Only relevant with `use_precomputed=True`.
        """
        assert self.signature == self.default_signature, 'The dataset arguments do not match the precomputed dataset arguments (the default settings). Set use_precomputed to False if you wish to generate a new dataset.'

    def proteins(self, resolution='residue'):
        """ Returns a generator of proteins from the avro file.

        Parameters
        ----------
        resolution: str, default 'residue'
            The resolution of the proteins. Can be 'atom' or 'residue'.

        Returns
        -------
        generator
            An avro reader object.


        .. code-block:: python

            >>> from proteinshake.datasets import RCSBDataset
            >>> protein = next(RCSBDataset().proteins())
        """
        self.download_precomputed(resolution=resolution)
        with open(f'{self.root}/{self.name}.{resolution}.avro', 'rb') as file:
            total = int(avro_reader(file).metadata['number_of_proteins'])
        def reader():
            with open(f'{self.root}/{self.name}.{resolution}.avro', 'rb') as file:
                for x in avro_reader(file):
                    yield x
        return Generator(reader(), total)

    @property
    def limit(self):
        """ Used only in testing, where this method is mock.patched to a small number. Default None.

        Returns
        -------
        int
            The limit to be applied to the number of downloaded/parsed files.
        """
        return None

    @property
    def name(self):
        return self.__class__.__name__

    def get_raw_files(self):
        """ Implement me in a subclass!

        Returns a list of all valid PDB file paths for this dataset. Usually takes the form `glob.glob(f'{self.root}/raw/files/*.pdb')` to search for all pdb files in the root, but can be different in some cases (e.g. with pdb.gz files).

        Returns
        -------
        list
            The list of raw PDB files used in this dataset.
        """
        raise NotImplementedError

    def get_id_from_filename(self, filename):
        """ Implement me in a subclass!

        Extracts an identifier from the pdb filename. This identifier is used in the `ID` field of the parsed protein object. Usually something like `filename[:4]`.

        Parameters
        ----------
        filename: str
            Path to a PDB file.

        Returns
        -------
        str
            A PDB identifier or other ID.
        """
        raise NotImplementedError

    def download(self):
        """ Implement me in a subclass!

        This method is responsible for downloading and extracting raw pdb files from a databank source. All PDB files should be dumped in `f'{self.root}/raw/files`. See e.g. PDBBindRefined for an example.
        """
        raise NotImplementedError

    def add_protein_attributes(self, protein):
        """ Implement me in a subclass!

        This method annotates protein objects with addititional information, such as functional labels or classes. It takes a protein object (a dictionary), modifies, and returns it. Usually, this would utilize the `ID` attribute to load an annotation file or to query information from a database.

        Parameters
        ----------
        protein: dict
            A protein object. See :meth:`proteinshake.datasets.Dataset.parse_pdb` for details.

        Returns
        -------
        dict
            The protein object with a new attribute added.
        """
        return protein

    def download_complete(self):
        """ Dumps a marker file when the download was successful, to skip downloading next time.
        """
        with open(f'{self.root}/raw/done.txt','w') as file:
            file.write('done.')

    def start_download(self):
        """ Helper function to prepare the download. Creates necessary subdirectories.
        """
        if os.path.exists(f'{self.root}/raw/done.txt'):
            return
        os.makedirs(f'{self.root}/raw/files', exist_ok=True)
        self.download()
        self.download_complete()

    def download_precomputed(self, resolution='residue'):
        """ Downloads the precomputed dataset from the ProteinShake repository.
        """
        if not os.path.exists(f'{self.root}/{self.name}.{resolution}.avro'):
            download_url(f'{self.repository_url}/{self.name}.{resolution}.avro.gz', f'{self.root}')
            print('Unzipping...')
            unzip_file(f'{self.root}/{self.name}.{resolution}.avro.gz')

    def parse(self):
        """ Parses all PDB files returned from :meth:`proteinshake.datasets.Dataset.get_raw_files()` and saves them to disk. Can run in parallel.
        """
        if os.path.exists(f'{self.root}/{self.name}.residue.avro'):
            return
        # parse and filter
        paths = self.get_raw_files()[:self.limit]
        proteins = Parallel(n_jobs=self.n_jobs)(delayed(self.parse_pdb)(path) for path in tqdm(paths, desc='Parsing'))
        before = len(proteins)
        proteins = [p for p in proteins if p is not None]
        print(f'Filtered {before-len(proteins)} proteins.')
        residue_proteins = [{'protein':p['protein'], 'residue':p['residue']} for p in proteins]
        atom_proteins = [{'protein':p['protein'], 'atom':p['atom']} for p in proteins]
        write_avro(residue_proteins, f'{self.root}/{self.name}.residue.avro')
        write_avro(atom_proteins, f'{self.root}/{self.name}.atom.avro')

    def parse_pdb(self, path):
        """ Parses a single PDB file first into a DataFrame, then into a protein object (a dictionary). Also validates the PDB file and provides the hook for `add_protein_attributes`. Returns `None` if the protein was found to be invalid.
        Parameters
        ----------
        path: str
            Path to PDB file.
        Returns
        -------
        dict
            A protein object.
        """
        pdbid = self.get_id_from_filename(os.path.basename(path))
        if pdbid in self.exclude_ids:
            return None
        atom_df = self.pdb2df(path)
        residue_df = atom_df[atom_df['atom_type'] == 'CA']
        if not self.validate(atom_df):
            return None

        # add surface accessible area
        structure = freesasa.Structure(path)
        result = freesasa.calc(structure)
        residue_result = result.residueAreas()
        atom_sasa, residue_sasa, residue_rsa = [], [], []
        for i in atom_df['atom_number']:
            try:
                atom_sasa.append(result.atomArea(i))
            except:
                atom_sasa.append(-1)
        for i,chain in zip(residue_df['residue_number'], residue_df['chain_id']):
            try:
                residue_sasa.append(residue_result[chain][str(i)].total)
                residue_rsa.append(residue_result[chain][str(i)].relativeTotal)
            except:
                residue_sasa.append(-1)
                residue_rsa.append(-1)
            

        # create protein_dict
        protein = {
            'protein': {
                'ID': pdbid,
                'sequence': ''.join(residue_df['residue_type']),
            },
            'residue': {
                'residue_number': residue_df['residue_number'].tolist(),
                'residue_type': residue_df['residue_type'].tolist(),
                'x': residue_df['x'].tolist(),
                'y': residue_df['y'].tolist(),
                'z': residue_df['z'].tolist(),
                'SASA': residue_sasa,
                'RSA': residue_rsa,
            },
            'atom': {
                'atom_number': atom_df['atom_number'].tolist(),
                'atom_type': atom_df['atom_type'].tolist(),
                'residue_number': atom_df['residue_number'].tolist(),
                'residue_type': atom_df['residue_type'].tolist(),
                'x': atom_df['x'].tolist(),
                'y': atom_df['y'].tolist(),
                'z': atom_df['z'].tolist(),
                'SASA': atom_sasa,
            },
        }

        # only include chains if multi-chain protein
        if not self.only_single_chain: 
            protein['residue']['chain_id'] = residue_df['chain_id'].tolist()
            protein['atom']['chain_id'] = atom_df['chain_id'].tolist()

        # add pLDDT from AlphaFold
        if self.name == 'AlphaFoldDataset':
            protein['residue']['pLDDT'] = residue_df['b_factor'].tolist()
            protein['atom']['pLDDT'] = atom_df['b_factor'].tolist()

        # add attributes
        protein = self.add_protein_attributes(protein)

        return protein

    def pdb2df(self, path):
        """ Parses a single PDB file to a DataFrame (with biopandas). Also deals with multiple structure models in a PDB (e.g. from NMR) by only selecting the first model.

        Parameters
        ----------
        path: str
            Path to PDB file.

        Returns
        -------
        DataFrame
            A biopandas DataFrame of the PDB file.
        """
        with open(path, 'r') as file:
            lines = file.read().split('\n')
        # filter only the first model
        filtered_lines, in_model, model_done = [], False, False
        for line in lines:
            if line.startswith('MODEL'):
                in_model = True
            if in_model and model_done:
                continue
            if line.startswith('ENDMDL'):
                model_done = True
                in_model = False
            filtered_lines.append(line)
        IONS = ['ZN', 'MG']
        df = PandasPdb().read_pdb_from_list(filtered_lines).df['ATOM']
        df = df.loc[~df['residue_name'].isin(IONS)]
        df['residue_name'] = df['residue_name'].map(lambda x: AA_THREE_TO_ONE[x] if x in AA_THREE_TO_ONE else None)
        #df['atom_name'] = df['atom_name'].map(lambda x: x[0]) # each atom is a multi-letter code where the first letter indicates the atom type
        df = df.rename(columns={
            'atom_name': 'atom_type',
            'residue_name': 'residue_type',
            'x_coord': 'x',
            'y_coord': 'y',
            'z_coord': 'z',
        })
        df = df.sort_values(by=['chain_id', 'residue_number', 'atom_number'])
        return df

    def validate(self, df):
        """ Performs several sanity checks for validity of the protein DataFrame.

        Parameters
        ----------
        df: DataFrame
            A protein DataFrame.

        Returns
        -------
        bool
            Whether or not the DataFrame is valid.
        """
        if len(df['residue_number'].unique()) < self.minimum_length:
            return False
        if len(df['residue_number'].unique()) > self.maximum_length:
            return False
        # check if single chain protein
        if self.only_single_chain and len(df['chain_id'].unique()) > 1:
            return False
        # check if sequence and structure are consistent
        if self.check_sequence and not max(df['residue_number']) == len(df['residue_number'].unique()):
            return False
        # check if all standard amino acids
        if not all(df['residue_type'].map(lambda x: not x is None)):
            return False
        # check if sequence is empty (e.g. with non-canonical amino acids)
        if not sum(df['residue_type'].map(lambda x: not x is None)) > 0:
            return False
        return True

    def describe(self):
        """ Produces dataset statistics.

        Returns
        -------
        dict
            A dictionary of summary statistics of this dataset.
        """
        n_resi = len(self.data.residue_index) / len(self.data.ID)
        data = {'name': type(self).__name__,
                'num_proteins': len(self),
                'avg size (# residues)': n_resi
               }
        return data
    
    def to_graph(self, resolution='residue', transform=IdentityTransform(), *args, **kwargs):
        """ Converts the raw dataset to a graph dataset. See :meth:`proteinshake.representations.GraphDataset` for arguments.

        Returns
        -------
        proteinshake.representations.GraphDataset
            The dataset in graph representation.
        """
        from proteinshake.representations import GraphDataset
        proteins = self.proteins(resolution=resolution)
        return GraphDataset(Generator((transform(p) for p in proteins), len(proteins)),
                            self.root,
                            self.name,
                            resolution,
                            *args,
                            **kwargs)

    def to_point(self, resolution='residue', transform=IdentityTransform(), *args, **kwargs):
        """ Converts the raw dataset to a point cloud dataset. See :meth:`proteinshake.representations.PointDataset` for arguments.

        Returns
        -------
        proteinshake.representations.PointDataset
            The dataset in point cloud representation.
        """
        from proteinshake.representations import PointDataset
        proteins = self.proteins(resolution=resolution)
        return PointDataset(Generator((transform(p) for p in proteins), len(proteins)),
                            self.root,
                            self.name,
                            resolution,
                            *args,
                            **kwargs)

    def to_voxel(self, resolution='residue', transform=IdentityTransform(), *args, **kwargs):
        """ Converts the raw dataset to a voxel dataset. See :meth:`proteinshake.representations.VoxelDataset` for arguments.

        Returns
        -------
        proteinshake.representations.VoxelDataset
            The dataset in voxel representation.
        """
        from proteinshake.representations import VoxelDataset
        proteins = self.proteins(resolution=resolution)
        return VoxelDataset(Generator((transform(p) for p in proteins), len(proteins)),
                            self.root,
                            self.name,
                            resolution,
                            *args,
                            **kwargs)



import json

from proteinshake.datasets import RCSBDataset

class EnzymeCommissionDataset(RCSBDataset):
    """
    Datsaet of enzymes which are assigned label based
    on the reaction they catalyze.


    .. list-table:: Dataset stats
       :widths: 100
       :header-rows: 1

       * - # proteins
       * - 15603 


   .. list-table:: Annotations
      :widths: 25 35 45
      :header-rows: 1

      * - Attribute
        - Key
        - Sample value
      * - Enzyme Commission
        - :code:`protein['protein']['EC']`
        - :code:`'2.7.7.4'`

    """

    def __init__(self, query=[['rcsb_polymer_entity.rcsb_ec_lineage.name','exists']], **kwargs):
        """

        Args:
            query: REST-API query.

        """
        super().__init__(query=query, **kwargs)

    def add_protein_attributes(self, protein):
        """ Fetch the enzyme class for each protein.
        """
        with open(f'{self.root}/raw/files/{protein["protein"]["ID"]}.annot.json','r') as file:
            annot = json.load(file)
        protein['protein']['EC'] = annot['rcsb_polymer_entity']['rcsb_ec_lineage'][-1]['id']
        return protein



# -*- coding: utf-8 -*-
import glob
import os

from biopandas.pdb import PandasPdb
from unittest.mock import patch
from tqdm import tqdm

from proteinshake.datasets.dataset import Dataset, AA_THREE_TO_ONE
from proteinshake.utils import download_url

EXTENDED_AA_THREE_TO_ONE = {
    **AA_THREE_TO_ONE,
    'CYZ': 'C',
    'CYX': 'C',
    'HIP': 'H',
    'HID': 'H',
    'HIE': 'H',
}

class ProteinLigandDecoysDataset(Dataset):
    """ Proteins (targets) from `DUDE-Z <https://pubs.acs.org/doi/10.1021/acs.jcim.0c00598>`_ with a list of decoys and active molecules for each.
    Each molecule is encoded as a SMILES string, meant to be used in a virtual screen setting.
    In this setting a model is given a protein and a ligand and outputs a score reflecting the likelihood
    that the given molecule is a binder. Then, this score is used to sort the union of all the ligands and
    decoys. A good model places true ligands at the top of this list. This is known as enrichment factor
    analysis.
    `data source <https://dudez.docking.org/>`_.


    .. list-table:: Dataset stats
       :widths: 100
       :header-rows: 1

       * - # proteins
       * - 38


   .. list-table:: Annotations
      :widths: 25 35 45
      :header-rows: 1

      * - Attribute
        - Key
        - Sample value
      * - Non-binders SIMLES
        - :code:`protein['protein']['deocys_smiles']`
        - :code:`['O=C(CSc1nnc(COc2ccccc2)o1)NC1CCCCC1', 'C[N@H+]1CC[C@@](N)(C(=O)NC[C@@H]2CC[C@@H](C[NH3+])CC2)C1',..]`
      * - Non-binders identifiers
        - :code:`protein['protein']['decoys_ids']`
        - :code:`['ZINC000000087599', 'ZINC000648138664',..]`
      * - Binders SIMLES
        - :code:`protein['protein']['ligands_smiles']`
        - :code:`['CC1=CC2=C(NC(=O)[C@H](CC3CC3)C2)C(=O)N1CC(=O)NCC1=CC=C(N)N=C1C', 'ClC1=CC=CC(CC2=NC3=C(NCCC4CCCC[NH2+]4)N=CC=C3O2)=C1',..]`
      * - Binders identifiers 
        - :code:`protein['protein']['ligands_ids']`
        - :code:`['CHEMBL10785', 'CHEMBL439678', 'CHEMBL278985',..]`
      * - Pfam accession code
        - :code:`protein['protein']['Pfam']`
        - ``['PF00102']``


    """

    @patch('proteinshake.datasets.dataset.AA_THREE_TO_ONE', EXTENDED_AA_THREE_TO_ONE)
    def pdb2df(self, path):
        return super().pdb2df(path)

    def get_raw_files(self):
        return glob.glob(f'{self.root}/raw/files/*.pdb')[:self.limit]

    def get_id_from_filename(self, filename):
        return filename.split(".")[0]

    def download(self):
        targets  = ['AA2AR', 'ABL1', 'ACES', 'ADA', 'ADRB2', 'AMPC', 'ANDR', 'CSF1R', 'CXCR4', 'DEF', 'DRD4', 'EGFR', 'FA7', 'FA10', 'FABP4', 'FGFR1', 'FKB1A', 'GLCM', 'HDAC8', 'HIVPR', 'HMDH', 'HS90A', 'ITAL', 'KITH', 'KIT', 'LCK', 'MAPK2', 'MK01', 'MT1', 'NRAM', 'PARP1', 'PLK1', 'PPARA', 'PTN1', 'PUR2', 'RENI', 'ROCK1', 'SRC', 'THRB', 'TRY1', 'TRYB1', 'UROK', 'XIAP']

        for target_id in tqdm(targets, desc='Downloading'):
            # grab receptor
            download_url(f"https://dudez.docking.org/DOCKING_GRIDS_AND_POSES/{target_id}/rec.crg.pdb", f"{self.root}/raw/files/", log=False)
            os.rename(f'{self.root}/raw/files/rec.crg.pdb', f'{self.root}/raw/files/{target_id}.pdb')
            # grab ligands
            download_url(f"https://dudez.docking.org/property_matched/{target_id}_new_DUDE_1/ligands.smi", f"{self.root}/raw/files/", log=False)
            os.rename(f'{self.root}/raw/files/ligands.smi', f'{self.root}/raw/files/ligands_{target_id}.smi')
            # grab decoys
            download_url(f"https://dudez.docking.org/property_matched/{target_id}_new_DUDE_1/decoys.smi", f"{self.root}/raw/files/", log=False)
            os.rename(f'{self.root}/raw/files/decoys.smi', f'{self.root}/raw/files/decoys_{target_id}.smi')


    def add_protein_attributes(self, protein):
        """ We annotte each protein with a list of decoys and a list of active SMILES strings and molecule IDs.
        """
        target = protein['protein']['ID']
        
        for mode in ['decoys', 'ligands']:
            with open(f"{self.root}/raw/files/{mode}_{target}.smi", "r") as mols:
                smiles, ids = [], []
                for line in mols:
                    smile, mol_id = line.split()
                    smiles.append(smile)
                    ids.append(mol_id)

            protein['protein'][f'{mode}_smiles'] = smiles
            protein['protein'][f'{mode}_ids'] = ids
            protein['protein'][f'num_{mode}'] = len(ids)

        protein['protein']['num_mols'] = protein['protein']['num_ligands'] + protein['protein']['num_decoys']
        return protein




from sklearn import metrics

from proteinshake.datasets import ProteinLigandInterfaceDataset
from proteinshake.tasks import Task

class LigandAffinityTask(Task):
    """ Predict the dissociation constant (kd) for a protein and a small molecule. This is a protein-level regression task.
    Small molecule ligand information is stored as ``dataset[i].smiles`` for a SMILES string, or as pre-computed molecular
    fingerprints ``dataset[i].fp_maccs``, ```dataset[i].fp_morgan_r2``.
    """
    
    DatasetClass = ProteinLigandInterfaceDataset

    @property
    def task_type(self):
        return ('protein', 'regression')

    def dummy_output(self):
        import random
        return [random.random() for _ in range(len(self.test_index))] 

    def target(self, protein):
        return protein['protein']['neglog_aff']

    def evaluate(self, y_pred):
        return {
            'mse': metrics.mean_squared_error(self.test_targets, y_pred),
            'r2': metrics.r2_score(self.test_targets, y_pred)
        }



# -*- coding: utf-8 -*-
'''
TMalign needs to be in your $PATH. Follow the instructions at https://zhanggroup.org/TM-align/readme.c++.txt
'''
import glob
import requests
import os
import itertools
import re
import subprocess
import tempfile
import shutil
import numpy as np
from biopandas.pdb import PandasPdb
from collections import defaultdict
from joblib import Parallel, delayed
from tqdm import tqdm
from functools import cached_property

from proteinshake.datasets import RCSBDataset
from proteinshake.utils import (extract_tar,
                                download_url,
                                save,
                                load,
                                unzip_file,
                                global_distance_test,
                                local_distance_difference_test
                                )


class TMAlignDataset(RCSBDataset):
    """Proteins that were aligned with TMalign. The dataset provides the TM-score, RMSD, Global Distance Test (GDT), and Local Distance Difference Test (LDDT) as similarity/distance metrics between any two proteins.

    .. list-table:: Dataset stats
       :widths: 100
       :header-rows: 1

       * - # proteins
       * - 994


    .. code-block:: python

        from proteinshake.datasets import TMAlignDataset

        dataset = TMAlignDataset()
        proteins = dataset.proteins()
        protein_1, protein_2 = next(proteins)['protein']['ID'], next(proteins)['protein']['ID']

        dataset.tm_score(protein_1, protein_2)
        >>> 0.03
        dataset.rmsd(protein_1, protein_2)
        >>> 3.64
        dataset.gdt(protein_1, protein_2)
        >>> 0.61
        dataset.lddt(protein_1, protein_2)
        >>> 0.65

    """

    additional_files = [
        'TMAlignDataset.tmscore.npy',
        'TMAlignDataset.gdt.npy',
        'TMAlignDataset.rmsd.npy',
        'TMAlignDataset.lddt.npy'
    ]

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        if not self.use_precomputed: self.align_structures()
        self.protein_ids = [p['protein']['ID'] for p in self.proteins()]

        def download_file(filename):
            if not os.path.exists(f'{self.root}/{filename}'):
                download_url(f'{self.repository_url}/{filename}.gz', f'{self.root}', log=False)
                unzip_file(f'{self.root}/{filename}.gz')
            return load(f'{self.root}/{filename}')

        self._tm_score = download_file(f'{self.name}.tmscore.npy')
        self._rmsd = download_file(f'{self.name}.rmsd.npy')
        self._gdt = download_file(f'{self.name}.gdt.npy')
        self._lddt = download_file(f'{self.name}.lddt.npy')
        
    @property
    def limit(self):
        return 1000
    
    def align_structures(self):
        """ Calls TMAlignn on all pairs of structures and saves the output"""
        if os.path.exists(f'{self.root}/{self.name}.tmscore.npy'):
            return
        pdbids = [p['protein']['ID'] for p in self.proteins()]
        path_dict = {self.get_id_from_filename(os.path.basename(f)):f for f in self.get_raw_files()}
        paths = [path_dict[id] for id in pdbids]
        num_proteins = len(paths)
        combinations = np.array(list(itertools.combinations(range(num_proteins), 2)))
        TM, RMSD, GDT, LDDT = [np.ones((num_proteins,num_proteins), dtype=np.float16) * np.nan for _ in ['tm','rmsd','gdt','lddt']]
        np.fill_diagonal(TM, 1.0), np.fill_diagonal(RMSD, 0.0), np.fill_diagonal(GDT, 1.0), np.fill_diagonal(LDDT, 1.0)
        d = Parallel(n_jobs=self.n_jobs)(delayed(tmalign_wrapper)(paths[i], paths[j]) for i,j in tqdm(combinations, desc='Aligning'))
        x,y = tuple(combinations[:,0]), tuple(combinations[:,1])
        TM[x,y] = [x['TM1'] for x in d]
        TM[y,x] = [x['TM2'] for x in d]
        RMSD[x,y] = [x['RMSD'] for x in d]
        RMSD[y,x] = [x['RMSD'] for x in d]
        GDT[x,y] = [x['GDT'] for x in d]
        GDT[y,x] = [x['GDT'] for x in d]
        LDDT[x,y] = [x['LDDT'] for x in d]
        LDDT[y,x] = [x['LDDT'] for x in d]
        # save
        np.save(f'{self.root}/{self.name}.tmscore.npy', TM)
        np.save(f'{self.root}/{self.name}.rmsd.npy', RMSD)
        np.save(f'{self.root}/{self.name}.gdt.npy', GDT)
        np.save(f'{self.root}/{self.name}.lddt.npy', LDDT)

    def tm_score(self, protein_1, protein_2):
        return self._tm_score[self.protein_ids.index(protein_1)][self.protein_ids.index(protein_2)]
    
    def rmsd(self, protein_1, protein_2):
        return self._rmsd[self.protein_ids.index(protein_1)][self.protein_ids.index(protein_2)]

    def gdt(self, protein_1, protein_2):
        return self._gdt[self.protein_ids.index(protein_1)][self.protein_ids.index(protein_2)]

    def lddt(self, protein_1, protein_2):
        return self._lddt[self.protein_ids.index(protein_1)][self.protein_ids.index(protein_2)]

def tmalign_wrapper(pdb1, pdb2):
    """Compute TM score with TMalign between two PDB structures.
    Parameters
    ----------
    pdb1: str
        Path to PDB.
    pdb2 : str
        Path to PDB.
    return_superposition: bool
        If True, returns a protein dataframe with superposed structures.
    Returns
    -------
    dict
        Metric values TM1/TM2 (TM-Scores normalized to pdb1 or pdb2), RMSD, GDT
    """
    assert shutil.which('TMalign') is not None,\
           "No TMalign installation found. Go here to install : https://zhanggroup.org/TM-align/TMalign.cpp"
    with tempfile.TemporaryDirectory() as tmpdir:
        lines = subprocess.run(['TMalign','-outfmt','-1', pdb1, pdb2, '-o', f'{tmpdir}/superposition'], stdout=subprocess.PIPE).stdout.decode().split('\n')
        TM1 = lines[7].split()[1]
        TM2 = lines[8].split()[1]
        RMSD = lines[6].split()[4][:-1]
        seq1, ali, seq2 = lines[12], lines[13], lines[14]
        i, j, alignmentA, alignmentB = 0, 0, [], []
        for s1,a,s2 in zip(seq1,ali,seq2):
            if a != ' ': alignmentA.append(i)
            if a != ' ': alignmentB.append(j)
            if s1 != '-': i += 1
            if s2 != '-': j += 1
        os.rename(f'{tmpdir}/superposition_all', f'{tmpdir}/superposition_all.pdb')
        superposition = PandasPdb().read_pdb(f'{tmpdir}/superposition_all.pdb')
        df = superposition.df['ATOM']
        A = df[df['chain_id'] == 'A']
        B = df[df['chain_id'] == 'B']
        coordsA = np.array(list(zip(A['x_coord'], A['y_coord'], A['z_coord'])))[alignmentA]
        coordsB = np.array(list(zip(B['x_coord'], B['y_coord'], B['z_coord'])))[alignmentB]
        GDT = global_distance_test(coordsA, coordsB)
        LDDT = local_distance_difference_test(coordsA, coordsB)
    return {
        'TM1': float(TM1),
        'TM2': float(TM2),
        'RMSD': float(RMSD),
        'GDT': GDT,
        'LDDT': LDDT
    }



import numpy as np
from scipy.spatial.distance import pdist, squareform

def global_distance_test(coordsA, coordsB):
    """
    Computes GDT. Input: aligned coordinates of both structures.
    """
    dist = np.sqrt(((coordsA-coordsB)**2).sum(1))
    thresholds = [1,2,4,8]
    return np.mean([np.mean(dist<=t) for t in thresholds])

def local_distance_difference_test(coordsA, coordsB):
    # parameters from 
    R0 = 15
    thresholds = [0.5,1,2,4]

    distA = squareform(pdist(coordsA))
    np.fill_diagonal(distA, np.nan)
    distA[distA>R0] = np.nan

    distB = squareform(pdist(coordsB))
    np.fill_diagonal(distB, np.nan)
    distB[distB>R0] = np.nan

    diff = np.abs(distA.reshape(-1) - distB.reshape(-1))
    return np.mean([np.sum(diff<=t)/np.isfinite(diff).sum() for t in thresholds])





'''
Tests loading a dataset into different frameworks and the specific data loaders.
'''

import unittest, tempfile
from proteinshake.datasets import ProteinLigandDecoysDataset

class TestFrameworks(unittest.TestCase):

    @classmethod
    def setUpClass(self):
        self.tmpdir = tempfile.TemporaryDirectory()
        self.ds = ProteinLigandDecoysDataset(root=self.tmpdir)

    @classmethod
    def tearDownClass(self):
        self.tmpdir.cleanup()

    def test_graph_pyg(self):
        from torch_geometric.loader import DataLoader
        graphs = self.ds.to_graph(k=5).pyg()
        loader = DataLoader(graphs)
        x = next(iter(loader))

    def test_graph_dgl(self):
        from dgl.dataloading import GraphDataLoader as DataLoader
        graphs = self.ds.to_graph(k=5).dgl()
        loader = DataLoader(graphs)
        x = next(iter(loader))

    def test_graph_nx(self):
        graphs = self.ds.to_graph(k=5).nx()
        x = graphs[0]

    def test_voxel_torch(self):
        from torch.utils.data import DataLoader
        voxels = self.ds.to_voxel().torch()
        loader = DataLoader(voxels)
        x = next(iter(loader))

    def test_voxel_tf(self):
        import tensorflow as tf
        voxels = self.ds.to_voxel().tf()
        def generator():
            for data, protein_dict in voxels:
                yield data
        loader = tf.data.Dataset.from_generator(generator, output_types=(tf.float32))
        x = next(iter(loader))

    def test_voxel_np(self):
        import numpy as np
        points = self.ds.to_point().np()
        def generator():
            for data, protein_dict in points:
                yield data
        loader = np.fromiter(generator(), object)
        x = next(iter(loader))

    def test_point_torch(self):
        from torch.utils.data import DataLoader
        points = self.ds.to_point().torch()
        loader = DataLoader(points)
        x = next(iter(loader))

    def test_point_tf(self):
        import tensorflow as tf
        points = self.ds.to_point().tf()
        def generator():
            for data, protein_dict in points:
                yield data
        loader = tf.data.Dataset.from_generator(generator, output_types=(tf.float32))
        x = next(iter(loader))

    def test_point_np(self):
        import numpy as np
        points = self.ds.to_point().np()
        def generator():
            for data, protein_dict in points:
                yield data
        loader = np.fromiter(generator(), object)
        x = next(iter(loader))


if __name__ == '__main__':
    unittest.main()



'''
Tests downloading the precomputed datasets and loading the data.
'''

import unittest, tempfile
from proteinshake.datasets import *
from proteinshake.datasets.alphafold import AF_DATASET_NAMES

class TestDatasets(unittest.TestCase):

    def test_pli(self):
        with tempfile.TemporaryDirectory() as tmp:
            ds = ProteinLigandInterfaceDataset(root=tmp).download_precomputed()

    def test_ppi(self):
        with tempfile.TemporaryDirectory() as tmp:
            ds = ProteinProteinInterfaceDataset(root=tmp).download_precomputed()

    def test_tm(self):
        with tempfile.TemporaryDirectory() as tmp:
            ds = TMAlignDataset(root=tmp).download_precomputed()

    def test_go(self):
        with tempfile.TemporaryDirectory() as tmp:
            ds = GeneOntologyDataset(root=tmp).download_precomputed()

    def test_ec(self):
        with tempfile.TemporaryDirectory() as tmp:
            ds = EnzymeCommissionDataset(root=tmp).download_precomputed()

    def test_pfam(self):
        with tempfile.TemporaryDirectory() as tmp:
            ds = ProteinFamilyDataset(root=tmp).download_precomputed()

    def test_rcsb(self):
        with tempfile.TemporaryDirectory() as tmp:
            ds = RCSBDataset(root=tmp).download_precomputed()

    def test_af(self):
        organism = 'methanocaldococcus jannaschii'
        with tempfile.TemporaryDirectory() as tmp:
            ds = AlphaFoldDataset(root=tmp, organism=organism).download_precomputed()

    def test_dude(self):
        with tempfile.TemporaryDirectory() as tmp:
            ds = ProteinLigandDecoysDataset(root=tmp).download_precomputed()

if __name__ == '__main__':
    unittest.main()



import torch
from torch_geometric.utils import from_scipy_sparse_matrix
from torch_geometric.data import Data, Dataset as PygDataset
from proteinshake.frameworks.dataset import FrameworkDataset


class PygGraphDataset(FrameworkDataset, PygDataset):

    def convert_to_framework(self, data_item):
        nodes, adj = data_item.data
        return Data(
            x = torch.from_numpy(nodes),
            edge_index = from_scipy_sparse_matrix(adj)[0].long(),
            edge_attr = from_scipy_sparse_matrix(adj)[1].unsqueeze(1).float()
        )



import unittest, tempfile

from proteinshake.datasets import ProteinLigandDecoysDataset

from proteinshake.transforms import *


class TestTransforms(unittest.TestCase):

    @classmethod
    def setUpClass(self):
        self.tmpdir = tempfile.TemporaryDirectory()
        self.ds = ProteinLigandDecoysDataset(root=self.tmpdir)

    @classmethod
    def tearDownClass(self):
        self.tmpdir.cleanup()

    def test_identity(self):
        self.ds.to_voxel(transform=IdentityTransform())

    def test_center(self):
        self.ds.to_voxel(transform=CenterTransform())

    def test_rotate(self):
        self.ds.to_voxel(transform=RandomRotateTransform())

    def test_compose(self):
        self.ds.to_voxel(transform=Compose([CenterTransform(), RandomRotateTransform()]))



from .task import Task
from .enzyme_class import EnzymeClassTask
from .pfam_task import ProteinFamilyTask
from .ligand_affinity import LigandAffinityTask
from .binding_site_detection import BindingSiteDetectionTask
from .structure_similarity import StructureSimilarityTask
from .protein_protein_interface import ProteinProteinInterfaceTask
from .structure_search import StructureSearchTask
from .structural_class import StructuralClassTask
from .gene_ontology import GeneOntologyTask
from .virtual_screen import VirtualScreenTask

classes = ['Task',
           'EnzymeClassTask',
           'ProteinFamilyTask',
           'GeneOntologyTask',
           'LigandAffinityTask',
           'BindingSiteDetectionTask',
           'ProteinProteinInterfaceTask',
           'StructuralClassTask',
           'StructureSimilarityTask',
           'StructureSearchTask',
           'VirtualScreenTask'
           ]

__all__ = classes



"""
Helper functions for all input/output related things.
"""

import os
import tarfile
import pickle
import json
import gzip
import shutil
import requests
import re
import pandas as pd
import numpy as np
from pathlib import Path
from tqdm import tqdm
from fastavro import writer as avro_writer, reader as avro_reader, parse_schema as parse_avro_schema

AA_THREE_TO_ONE = {'ALA': 'A', 'CYS': 'C', 'ASP': 'D', 'GLU': 'E', 'PHE': 'F', 'GLY': 'G', 'HIS': 'H', 'ILE': 'I', 'LYS': 'K', 'LEU': 'L', 'MET': 'M', 'ASN': 'N', 'PRO': 'P', 'GLN': 'Q', 'ARG': 'R', 'SER': 'S', 'THR': 'T', 'VAL': 'V', 'TRP': 'W', 'TYR': 'Y'}
AA_ONE_TO_THREE = {v:k for k, v in AA_THREE_TO_ONE.items()}

class Generator(object):
    def __init__(self, generator, length):
        self.generator = generator
        self.length = length

    def __len__(self): 
        return self.length

    def __iter__(self):
        return self.generator

    def __next__(self):
        return next(self.generator)

def fx2str(fx):
    """ Converts a function to a string representation.

    Parameters
    ----------
    fx: function
        A function.

    Returns
    -------
    str
        The stringified function.
    """
    return re.sub('(<.*?)\\s.*(>)', r'\1\2', fx.__repr__())

def avro_schema_from_protein(protein):
    """ Guesses the avro schema from a dictionary.

    Parameters
    ----------
    protein: dict
        A protein dictionary.

    Returns
    -------
    schema
        An avro schema.
    """
    typedict = {'int':'int', 'float':'float', 'str':'string', 'bool':'boolean'}
    def field_spec(k,v):
        if type(v) == dict:
            return {'name':k, 'type':{'name':k, 'type':'record', 'fields': [field_spec(_k,_v) for _k,_v in v.items()]}}
        elif type(v) == list:
            return {'name':k, 'type':{'type': 'array', 'items': typedict[type(v[0]).__name__] if len(v)>0 else 'string'}}
        elif type(v).__name__ in typedict:
            return {'name':k, 'type': typedict[type(v).__name__]}
        else:
            raise TypeError(f"All fields in a protein object need to be either int, float, bool or string, not {type(v).__name__}")
    schema = {
        'name': 'Protein',
        'namespace': 'Dataset',
        'type': 'record',
        'fields': [field_spec(k,v) for k,v in protein.items()],
    }
    return parse_avro_schema(schema)

def write_avro(proteins, path):
    """ Writes a list of protein dictionaries to an avro file.

    Parameters
    ----------
    proteins: list
        The list of proteins.
    path:
        The path to the output file.
    """
    path = Path(path)
    schema = avro_schema_from_protein(proteins[0])
    with open(path, 'wb') as file:
        avro_writer(file, schema, proteins, metadata={'number_of_proteins':str(len(proteins))})

def save(obj, path):
    """ Saves an object to either pickle, json, or json.gz (determined by the extension in the file name).

    Parameters
    ----------
    obj:
        The object to be saved.
    path:
        The path to save the object.
    """
    if path.endswith('.json.gz'):
        with gzip.open(Path(path), 'w') as file:
            file.write(json.dumps(obj).encode('utf-8'))
    elif path.endswith('.json'):
        with open(Path(path),'w') as file:
            json.dump(obj, file)
    elif path.endswith('.npy'):
        np.save(path, obj)
    else:
        with open(Path(path), 'wb') as file:
            pickle.dump(obj, file, protocol=pickle.HIGHEST_PROTOCOL)

def load(path):
    """ Loads a pickle, json or json.gz file.

    Parameters
    ----------
    path:
        The path to be loaded.

    Returns
    -------
    object
        The loaded object.
    """
    if path.endswith('.json.gz'):
        with gzip.open(Path(path), 'r') as file:
            obj = json.loads(file.read().decode('utf-8'))
    elif path.endswith('.json'):
        with open(Path(path),'r') as file:
            obj = json.load(file)
    elif path.endswith('.npy'):
        obj = np.load(path)
    else:
        with open(Path(path), 'rb') as handle:
            obj = pickle.load(handle)
    return obj

def zip_file(path):
    """ Zips a file.

    Parameters
    ----------
    path:
        The path to the file.

    """
    with open(Path(path), 'rb') as f_in:
        with gzip.open(Path(path+'.gz'), 'wb') as f_out:
            f_out.writelines(f_in)
    return path+'.gz'

def unzip_file(path, remove=True):
    """ Unzips a .gz file.

    Parameters
    ----------
    path:
        The path to the .gz file.

    """
    assert path.endswith('.gz')
    with gzip.open(Path(path), 'rb') as f_in:
        with open(Path(path[:-3]), 'wb') as f_out:
            shutil.copyfileobj(f_in, f_out)
    if remove:
        os.remove(path)
    return path[:-3]



def download_url(url, out_path, log=True, chunk_size=10*1024*1024):
    """ Downloads a file from an url. If `out_path` is a directory, the file will be saved under the url basename.

    Parameters
    ----------
    url: str
        The url to be downloaded.
    out_path: str
        Path to save the downloaded file.
    log: bool, default True
        Whether to show a progress bar.
    chunk_size: int, default 10485760
        The chunk size of the download.

    """
    file_name = os.path.basename(url)
    if os.path.isdir(out_path) or out_path.endswith('/'):
        out_path += '/'+file_name
    r = requests.get(url, stream=True)
    r.raise_for_status()
    total = int(r.headers.get('content-length', 0))
    if log:
        print(f'Downloading {file_name}:')
        bar = tqdm(total=total, unit='iB', unit_scale=True, unit_divisor=chunk_size)
    out_path = Path(out_path)
    with open(out_path, 'wb') as file:
        for data in r.iter_content(chunk_size=chunk_size):
            size = file.write(data)
            if log:
                bar.update(size)
    if log:
        bar.close()

def extract_tar(tar_path, out_path, extract_members=False, strip=0):
    """ Extracts a tar file.

    Parameters
    ----------
    tar_path:
        The path to the tar file.
    out_path:
        The directory to extract to.
    extract_members: bool, default False
        If `True`, the tar file member will be directly extracted to `out_path`, instead of creating a subdirectory.
    strip: int, default 0
        Remove `strip` folder hierarchies from the path of the extracted file.
    """
    def get_members(file):
        for member in file.getmembers():
            parts = Path(member.path).parts
            member.path = Path(*parts[min(strip, len(parts)-1):])
            yield member

    out_path = Path(out_path)
    with tarfile.open(tar_path,'r') as file:
        len_members = len(file.getmembers())
        members = get_members(file)
        if extract_members:
            for member in tqdm(members, desc='Extracting', total=len_members):
                file.extract(member, out_path)
        else:
            file.extractall(out_path, members=tqdm(file, desc='Extracting', total=len_members))

def protein_to_pdb(protein, path):
    """ Write coordinate list from atom dict to a PDB file.

    Parameters
    ------------
    protein:
        protein data dictionary. must be at the 'atom' resolution.
    path:
        Path to write PDB file.


    """
    path = Path(path)
    # ATOM      1  N   PRO A   1       8.316  21.206  21.530  1.00 17.44           N
    try:
        df = pd.DataFrame(protein['atom'])
        mode = 'atom'
    except KeyError:
        df = pd.DataFrame(protein['residue'])
        mode = 'residue'

    df['residue_name_full'] = df['residue_type'].apply(lambda x : AA_ONE_TO_THREE[x])
    if 'chain_id' not in df.columns:
        df['chain_id'] = 'A'
    df['occupancy'] = [1.00] * len(df)
    df['temp'] = [20.00] * len(df)

    if mode == 'atom':
        df['element'] = df['atom_type'].apply(lambda x: x[:1])
    else:
        df['element'] = 'C'
        df['atom_number'] = df['residue_number']
        df['atom_type'] = 'CA'

    lines = []
    for row in df.itertuples():
        line = "{:6s}{:5d} {:^4s}{:1s}{:3s} {:1s}{:4d}{:1s}   " \
               "{:8.3f}{:8.3f}{:8.3f}{:6.2f}{:6.2f}          " \
               "{:>2s}{:2s}".format(
                                    'ATOM',
                                    row.atom_number,
                                    row.atom_type,
                                    ' ',
                                    row.residue_name_full,
                                    row.chain_id,
                                    row.residue_number,
                                    ' ',
                                    row.x,
                                    row.y,
                                    row.z,
                                    row.occupancy,
                                    row.temp,
                                    row.element,
                                    '  '
                                 )
        lines.append(line)
    with open(path, "w") as p:
        p.write("\n".join(lines))



import os
import itertools
from tqdm import tqdm
import numpy as np

from proteinshake.utils import onehot

class Voxel():
    """ Voxel representation of a protein.

    Voxelizes a protein.

    Parameters
    ----------
    protein: dict
        A protein object.
    voxelsize: float
        The size of a voxel (in Angstrom).
    aggregation: str, defaul 'mean'
        How to aggregate labels of a voxel.

    """

    def __init__(self, protein, gridsize, voxelsize, aggregation):
        resolution = 'atom' if 'atom' in protein else 'residue'
        self.protein_dict = protein
        self.resolution = resolution
        labels = onehot(protein[resolution][f'{resolution}_type'])
        coords = np.stack([protein[resolution]['x'], protein[resolution]['y'], protein[resolution]['z']], axis=1)
        coords -= coords.min(axis=0) # translate to make all coords positive and flushed to the axes
        voxel_indices = (coords / voxelsize).astype(np.int32) # rasterize
        voxels = np.zeros(shape=np.concatenate([voxel_indices.max(axis=0)+1, labels.shape]))
        counts = np.zeros(shape=np.concatenate([voxel_indices.max(axis=0)+1, labels.shape]))
        voxel_indices = np.concatenate([voxel_indices, np.expand_dims(np.arange(len(coords)),1)], axis=1)
        voxels[tuple(voxel_indices.transpose())] = labels
        if aggregation == 'sum':
            voxels = voxels.sum(axis=-2)
        elif aggregation == 'mean':
            counts[tuple(voxel_indices.transpose())] = np.ones_like(labels)
            counts = counts.sum(axis=-2)
            voxels = np.divide(voxels.sum(axis=-2), counts, out=np.zeros_like(counts), where=counts!=0)
        # trim to gridsize
        diff = gridsize-voxels.shape[:-1]
        lower = -((diff < 0) * np.ceil(diff/2)).astype(int)
        upper = (voxels.shape[:-1] * (diff >= 0) + (diff < 0) * np.floor(diff/2)).astype(int)
        voxels = voxels[lower[0]:upper[0],lower[1]:upper[1],lower[2]:upper[2]]
        # pad to gridsize
        diff = gridsize-voxels.shape[:-1]
        paddings = np.stack([np.ceil(diff/2), np.floor(diff/2)],1).astype(np.int32)
        voxels = np.pad(voxels, (*paddings,(0,0)))
        self.data = voxels



class VoxelDataset():
    """ Voxel representation of a protein structure dataset.

    Parameters
    ----------
    proteins: generator
        A generator of protein objects from a Dataset.
    size: int
        The size of the dataset.
    path: str
        Path to save the processed dataset.
    resolution: str, default 'residue'
        Resolution of the proteins to use in the graph representation. Can be 'atom' or 'residue'.
    gridsize: tuple, default None
        The size of the grid in voxels as a 3-tuple of x,y,z edge lengths. If None (default), the dimensions of the largest protein in the dataset is used.
    voxelsize: float, default 10
        The size of a voxel (in Angstrom).
    aggregation: str, defaul 'mean'
        How to aggregate labels of a voxel.

    """

    def __init__(self, proteins, root, name, resolution='residue', gridsize=None, voxelsize=10, aggregation='mean'):
        self.size = len(proteins)
        if gridsize is None:
            proteins, proteins_copy = itertools.tee(proteins)
            gridsize = np.array([[
                np.ptp(protein[resolution]['x']),
                np.ptp(protein[resolution]['y']),
                np.ptp(protein[resolution]['z'])
                ] for protein in proteins_copy]).max(0)
            gridsize = np.ceil(gridsize/voxelsize).astype(int)
        gridsize = np.array(gridsize)
        gridsize_string = '_'.join(str(i) for i in gridsize)
        self.gridsize = gridsize
        self.voxels = (Voxel(protein, gridsize, voxelsize, aggregation) for protein in proteins)
        self.path = f'{root}/processed/voxel/{name}_{resolution}_voxelsize_{voxelsize}_gridsize_{gridsize_string}_aggregation_{aggregation}'

    def torch(self, *args, **kwargs):
        from proteinshake.frameworks.torch import TorchVoxelDataset
        return TorchVoxelDataset(self.voxels, self.size, self.path+'.torch', *args, **kwargs)

    def tf(self, *args, **kwargs):
        from proteinshake.frameworks.tf import TensorflowVoxelDataset
        return TensorflowVoxelDataset(self.voxels, self.size, self.path+'.tf', *args, **kwargs)

    def np(self, *args, **kwargs):
        from proteinshake.frameworks.np import NumpyVoxelDataset
        return NumpyVoxelDataset(self.voxels, self.size, self.path+".np", *args, **kwargs)



import numpy as np
from sklearn.model_selection import train_test_split

from proteinshake.datasets import ProteinLigandDecoysDataset
from proteinshake.tasks import Task

class VirtualScreenTask(Task):
    """ Test an affinity scoring model on a virtual screen.
    The goal in a virtual screen is: for a given protein and a library of potential
    binders, bring the binders to the top of the list.
    In this task, the model is given a protein and a list of ligands to score.
    The model scores each ligand in a library with a score proportional to the likelihood
    that the protein and ligand will bind. This can be a docking score, energy calculation,
    or just a probability.
    Each protein's ligand library contains a certain number of active molecules (ligands)
    and a certai (larger) number of decoys (non-binders).
    We use the predicted scores to sort the whole library and calculate the position of each
    active ligand in the sorted library.
    Ligands in the topi percentiles which are known to be active contribute a 1 to the score
    and those below the cutoff contribute a 0.

    .. warning::

        This is a zero-shot task so we use the whole dataset in
        evaluation. No train/test split.

    .. code-block:: python

        >>> from proteinshake.tasks import VirtualScreenTask
        >>> import numpy as np
        >>> task = VirtualScreenTask()
        # predict a (random) binding score for each molecule 
        >>> preds = [np.random.rand(len(task.target(p))) for p in task.dataset.proteins()]
        >>> task.evaluate(preds, cutoff_fraction=.2)
        {'enrichment_factor-@.2': 0.6}

    """

    DatasetClass = ProteinLigandDecoysDataset

    def __init__(self, *args, **kwargs):
        kwargs['split'] = 'none'
        super().__init__(*args, **kwargs)

    @property
    def task_type(self):
        return ('protein', 'virtual_screen')

    def target(self, protein):
        """ The target here is a sorted list of smiles where the true ligands
        occupy the top of the list and the decoys occupy the rest.
        Since this is a zero-shot task we only use this internally for evaluation.

        """
        return protein['protein']['ligands_smiles'] + protein['protein']['decoys_smiles']

    @property
    def num_features(self):
        return 20

    def dummy_output(self):
        import random
        return [[random.random() for _ in range(len(self.target(p)))] for p in self.proteins]

    def evaluate(self, y_pred, cutoff_fraction=.2):
        """ Computing enrichment factor on the whole dataset.
        To compute the EF, for each active ligand we check if it is in the top
        ``cutoff_fraction`` of the scored ligands and if so, count it as a 1 and 0 otherwise.
        Taking the mean gives a score between 0 and 1 where 1 is a perfect score.

        Arguments
        -----------

        y_pred: list[list[float]]
            A list of binding scores for each molecule returned by ``self.target(protein)``.
            We assume that a large value for the score means stronger likelihood of binding.
        cutoff_fraction: float
            Top fraction based on given scores within which to count actives.


        Returns
        ----------
        metrics: dict
            A dictionary with the enrichment factor.

        """
        
        efs = []
        for i, protein in enumerate(self.proteins):
            lig_ids = self.target(protein)
            active_ids = lig_ids[:protein['protein']['num_ligands']]

            cutoff_index = int(len(lig_ids) * cutoff_fraction)

            scores_dict = {lig_ids[i]: score for i, score in enumerate(y_pred[i])}
            ranks_dict = {lig_id: i < cutoff_index for i, lig_id in  enumerate(sorted(lig_ids, key = lambda x: scores_dict[x]))}

            mean_active_rank = np.mean([ranks_dict[lig_id] for lig_id in active_ids])
            efs.append(mean_active_rank)

        return {'enrichment_factor-@{self.cutoff_fraction}': np.mean(efs)}




# -*- coding: utf-8 -*-
import os
import glob
import pandas as pd
from biopandas.pdb import PandasPdb
from sklearn.neighbors import KDTree
import numpy as np

from proteinshake.datasets import Dataset
from proteinshake.utils import extract_tar, download_url

class ProteinProteinInterfaceDataset(Dataset):
    """Protein-protein complexes from PDBBind with annotated interfaces. Residues
    and atoms in each protein are marked with a boolean `is_interface` to indicate
    residues/atoms defined to belong to the interface of two protein chains.
    The default threshold for determining interface residues is 6 Angstroms.
    See :meth:`proteinshake.utils.get_interfaces` for details.

    Parameters
    ----------
    root: str
        Root directory where the dataset should be saved.
    name: str
        The name of the dataset.
    version: str
        PDBBind database version to use.
    cutoff: float
        Distance in angstroms within which a pair of residues is considered to
        belong to the interface.

    .. list-table:: Dataset stats
       :widths: 100
       :header-rows: 1

       * - # proteins
       * - 2839


   .. list-table:: Annotations
      :widths: 25 55 20
      :header-rows: 1

      * - Attribute
        - Key
        - Sample value
      * - Chain Identifier 
        - :code:`protein[{'residue' | 'atom'}]['chain_id']`
        - ``['X', 'X', ..'Y', 'Y']``
      * - Binding interface
        - :code:`protein[{'residue' | 'atom'}]['is_interface']`
        - ``[0, 0, .., 1, 0]``


    """


    def __init__(self, cutoff=6, version='2020', **kwargs):
        self.version = version
        self.cutoff = cutoff
        super().__init__(**kwargs)

    def get_raw_files(self):
        return glob.glob(f'{self.root}/raw/files/*.pdb')[:self.limit]

    def get_id_from_filename(self, filename):
        return filename[:4]

    def get_interfaces(self, protein, cutoff=6, resolution='residue'):
        """Obtain interfacing residues within a single structure of polymers. Uses
        KDTree data structure for vector search.

        Parameters
        ----------
        protein: dict
            Parsed protein dictionary.

        Returns
        --------
            `list`: indicator list for each residue with 0 if not in interface and 1 else.
        """

        #3-D KD tree
        df = pd.DataFrame({
            'x': protein[resolution]['x'],
            'y': protein[resolution]['y'],
            'z': protein[resolution]['z'],
            'chain': protein[resolution]['chain_id'],
            'residue_index':protein[resolution]['residue_number']
        })
        resi_df = df.groupby(['residue_index', 'chain']).mean().reset_index()
        resi_coords = np.array([resi_df['x'].tolist(), resi_df['y'].tolist(), resi_df['z'].tolist()]).T
        kdt = KDTree(resi_coords, leaf_size=1)

        query = kdt.query_radius(resi_coords, cutoff)
        interface = set()
        for i,result in enumerate(query):
            res_index = resi_df.iloc[i].name
            this_chain = resi_df.iloc[i].chain
            for r in result:
                that_resi = resi_df.iloc[r].name
                that_chain = resi_df.iloc[r].chain
                if this_chain != that_chain:
                    interface.add(res_index)
                    interface.add(that_resi)

        resi_interface = []
        for r in protein[resolution]['residue_number']:
            resi_interface.append(r in interface)
        return resi_interface

    def download(self):
        download_url(f'https://pdbbind.oss-cn-hangzhou.aliyuncs.com/download/PDBbind_v{self.version}_PP.tar.gz', f'{self.root}/raw')
        extract_tar(f'{self.root}/raw/PDBbind_v{self.version}_PP.tar.gz', f'{self.root}/raw/files', extract_members=True)

    def add_protein_attributes(self, protein):
        interface_atoms = self.get_interfaces(protein, self.cutoff, resolution='atom')
        protein['atom']['is_interface'] = interface_atoms
        protein['residue']['is_interface'] = [
            val
            for val, atom_type in \
            zip(interface_atoms, protein['atom']['atom_type']) \
            if atom_type == 'CA'
        ]
        return protein

    def describe(self):
        desc = super().describe()
        desc['property'] = "Protein-protein interface (residue-level)"
        desc['values'] = 2
        desc['type'] = "Binary"
        return desc



from sklearn import metrics

from proteinshake.datasets import ProteinLigandInterfaceDataset
from proteinshake.tasks import Task

class BindingSiteDetectionTask(Task):
    """ Identify the binding residues of a protein-small molecule binding site. This is a residue-level binary classification task.

    """
    
    DatasetClass = ProteinLigandInterfaceDataset

    @property
    def task_type(self):
        return ('residue', 'binary')

    def dummy_output(self):
        import random
        return [random.randint(0, 1) for p in self.test_targets]

    def target(self, protein):
        return protein['residue']['binding_site']

    def compute_targets(self):
        # compute targets (e.g. for scaling)
        self.train_targets = [p for i in self.train_index for p in self.target(self.proteins[i])]
        self.val_targets = [p for i in self.val_index for p in self.target(self.proteins[i])]
        self.test_targets = [p for i in self.test_index for p in self.target(self.proteins[i])]

    def evaluate(self, y_pred):
        return {
                'accuracy': metrics.accuracy_score(self.test_targets, y_pred),
                'mcc': metrics.matthews_corrcoef(self.test_targets, y_pred),
                }



# -*- coding: utf-8 -*-

import os

import numpy as np
from tqdm import tqdm
from scipy import sparse

from proteinshake.utils import load, save
from proteinshake.frameworks.dataset import FrameworkDataset

class NumpyVoxelDataset(FrameworkDataset):
    def convert_to_framework(self, data_item):
        return sparse.csr_array(data_item.data)

    def load_transform(self, data, protein_dict):
        return data.todense(), protein_dict


class NumpyPointDataset(FrameworkDataset):
    def convert_to_framework(self, data_item):
        return data_item.data

    def load_transform(self, data, protein_dict):
        return data, protein_dict



import os
from sklearn.neighbors import kneighbors_graph, radius_neighbors_graph
from tqdm import tqdm
import numpy as np

from proteinshake.utils import tokenize

class Graph():
    """ Graph representation of a protein.

    Converts a protein object to a graph by using a k-nearest-neighbor or epsilon-neighborhood approach. Define either `k` or `eps` to determine which one is used.

    Parameters
    ----------
    protein: dict
        A protein object.
    construction: str
        Whether to use knn or eps construction.
    eps: float
        The epsilon radius to be used in graph construction (in Angstrom).
    k: int
        The number of neighbors to be used in the k-NN graph.
    weighted_edges: bool, default False
        If `True`, edges are attributed with their euclidean distance. If `False`, edges are unweighted.

    """

    def __init__(self, protein, construction, k, eps, weighted_edges):
        resolution = 'atom' if 'atom' in protein else 'residue'
        mode = 'distance' if weighted_edges else 'connectivity'
        coords = np.stack([protein[resolution]['x'], protein[resolution]['y'], protein[resolution]['z']], axis=1)
        nodes = tokenize(protein[resolution][f'{resolution}_type'], resolution=resolution)
        if construction == 'eps':
            adj = radius_neighbors_graph(coords, radius=eps, mode=mode)
        elif construction == 'knn':
            n_neighbors = min(len(coords) - 1, k) # reduce k if protein is smaller than self.k
            adj = kneighbors_graph(coords,  n_neighbors=n_neighbors, mode=mode)
        self.protein_dict = protein
        self.resolution = resolution
        self.data = (nodes, adj)
        self.weighted_edges = weighted_edges



class GraphDataset():
    """ Graph representation of a protein structure dataset.

    Parameters
    ----------
    proteins: generator
        A generator of protein objects from a Dataset.
    size: int
        The size of the dataset.
    path: str
        Path to save the processed dataset.
    resolution: str, default 'residue'
        Resolution of the proteins to use in the graph representation. Can be 'atom' or 'residue'.
    eps: float
        The epsilon radius to be used in graph construction (in Angstrom).
    k: int
        The number of neighbors to be used in the k-NN graph.
    weighted_edges: bool, default False
        If `True`, edges are attributed with their euclidean distance. If `False`, edges are unweighted.

    """

    def __init__(self, proteins, root, name, resolution='residue', eps=None, k=None, weighted_edges=False):
        assert not (eps is None and k is None), 'You must specify eps or k in the graph construction.'
        construction = 'knn' if not k is None else 'eps'
        param = k if construction == 'knn' else eps
        weighted = '_weighted' if weighted_edges else ''
        self.path = f'{root}/processed/graph/{name}_{resolution}_{construction}_{param}{weighted}'
        self.graphs = (Graph(protein, construction, k, eps, weighted_edges) for protein in proteins)
        self.size = len(proteins)
        os.makedirs(os.path.dirname(self.path), exist_ok=True)

    def pyg(self, *args, **kwargs):
        from proteinshake.frameworks.pyg import PygGraphDataset
        return PygGraphDataset(self.graphs, self.size, self.path+'.pyg', *args, **kwargs)

    def dgl(self, *args, **kwargs):
        from proteinshake.frameworks.dgl import DGLGraphDataset
        return DGLGraphDataset(self.graphs, self.size, self.path+'.dgl', *args, **kwargs)

    def nx(self, *args, **kwargs):
        from proteinshake.frameworks.nx import NetworkxGraphDataset
        return NetworkxGraphDataset(self.graphs, self.size, self.path+'.nx', *args, **kwargs)



import itertools
import numpy as np
from sklearn import metrics
from functools import cached_property

from proteinshake.datasets import GeneOntologyDataset
from proteinshake.tasks import Task

class GeneOntologyTask(Task):
    """ Predict the Gene Ontology terms of a protein structure. This is a protein-level multi-label prediction.

    The prediction should be a n_samples x n_classes matrix, where the columns are ordered according to `self.classes`.
    If your model does not predict or handle a certain class, assign a zero value.
    """

    DatasetClass = GeneOntologyDataset
    
    def __init__(self, branch='molecular_function', *args, **kwargs):
        self.branch = branch
        super().__init__(*args, **kwargs)

    @cached_property
    def token_map(self):
        labels = set(itertools.chain(*[p['protein'][self.branch] for p in self.proteins]))
        return {label: i for i, label in enumerate(sorted(list(labels)))}

    @property
    def num_classes(self):
        return len(self.token_map)
    
    @property
    def classes(self):
        return list(self.token_map.keys())

    @property
    def task_type(self):
        return ('protein', 'multi_label')

    @property
    def num_features(self):
        return 20

    def target(self, protein):
        return [term in protein['protein'][self.branch] for term in self.classes]

    def precision(self, y_pred, threshold):
        y_true = np.copy(self.test_targets)[y_pred.max(axis=1) >= threshold]
        y_pred = y_pred[y_pred.max(axis=1) >= threshold] >= threshold
        mt = y_pred.shape[0]
        if mt == 0: return 0
        nom = np.logical_and(y_true, y_pred).sum(axis=1)
        denom = y_pred.sum(axis=1)
        return 1/mt * np.divide(nom, denom, out=np.zeros_like(nom), where=denom!=0).sum()

    def recall(self, y_pred, threshold):
        y_true = np.copy(self.test_targets)
        y_pred = y_pred >= threshold
        ne = y_true.shape[0]
        if ne == 0: return 0
        nom = np.logical_and(y_true, y_pred).sum(axis=1)
        denom = y_true.sum(axis=1)
        return 1/ne * np.divide(nom, denom, out=np.zeros_like(nom), where=denom!=0).sum()
    
    def remaining_uncertainty(self, y_pred, threshold):
        pass

    def missing_information(self, y_pred, threshold):
        pass

    def fmax(self, y_pred):
        fmax = 0
        for t in np.linspace(0,1,21):
            prec, rec = self.precision(y_pred, t), self.recall(y_pred, t)
            if prec+rec == 0: continue
            f1 = (2 * prec * rec) / (prec + rec)
            fmax = max(fmax, f1)
        return fmax
    
    def smin(self, y_pred):
        return min([
            np.sqrt(
                self.remaining_uncertainty(y_pred, t) ** 2
                + self.missing_information(y_pred, t) ** 2
            )
            for t in np.linspace(0,1,21)
        ])

    def dummy_output(self):
        import numpy as np
        return np.random.rand(len(self.test_index), len(self.token_map.keys()))

    def evaluate(self, y_pred):
        return {
            'Fmax': self.fmax(y_pred),
            #'Smin': self.smin(y_pred),
        }



import torch
from torch.utils.data import Dataset as TorchDataset
from proteinshake.frameworks.dataset import FrameworkDataset


class TorchVoxelDataset(FrameworkDataset, TorchDataset):

    def convert_to_framework(self, data_item):
        return torch.tensor(data_item.data).float().to_sparse()

    def load_transform(self, data, protein_dict):
        return data.to_dense(), protein_dict


class TorchPointDataset(FrameworkDataset, TorchDataset):

    def convert_to_framework(self, data_item):
        return torch.tensor(data_item.data).float()



from setuptools import find_packages, setup

__version__ = '0.3.7'
URL = 'https://proteinshake.readthedocs.io/en/latest/index.html'

install_requires = [
    'biopandas>=0.4.1',
    'pandas>=1.4.3',
    'rdkit-pypi>=2022.3.3',
    'tqdm>=4.64.0',
    'scikit-learn>=1.1.1',
    'joblib>=1.2.0',
    'requests>=2.27.1',
    'fastavro>=1.6.1',
    'freesasa>=2.2.0.post3',
    'goatools>=1.3.1',
]
test_requires = [
    'pytest',
]

setup(
    name='proteinshake',
    version=__version__,
    description='Deep learning ready datasets of 3D protein structures.',
    author = "Tim Kucera, Carlos Oliver, Dexiong Chen, Karsten Borgwardt",
    author_email = "kucera@biochem.mpg.de",
    url=URL,
    keywords = ['bioinformatics',
                'deep-learning',
                'computational-biology',
                'macromolecular-structure'],
    python_requires='>=3.7',
    install_requires=install_requires,
    packages=find_packages(),
    include_package_data=True,
)



import dgl
import torch
from dgl.data import DGLDataset
from proteinshake.frameworks.dataset import FrameworkDataset


class DGLGraphDataset(FrameworkDataset, DGLDataset):

    def convert_to_framework(self, data_item):
        nodes, adj = data_item.data
        data = dgl.from_scipy(adj, eweight_name='edge_weight')
        if data_item.weighted_edges:
            data.ndata[f'{data_item.resolution}'] = torch.tensor(nodes).long()
        return data



from .embeddings import *
from .io import *
from .similarity import *

__all__ = ['onehot',
           'tokenize',
           'positional_encoding',
           'compose_embeddings',
           'save',
           'load',
           'download_url',
           'extract_tar',
           'zip_file',
           'unzip_file',
           'write_avro',
           ]

classes = __all__
